import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from dash.dependencies import Output, Input, State
from sqlalchemy.orm import Session

from covigator.dashboard.figures.mutation_stats import MutationStatsFigures
from covigator.dashboard.figures.samples import SampleFigures
from covigator.database.model import DataSource, VariantType
from covigator.database.queries import Queries


ID_VARIANTS_PER_SAMPLE_GRAPH = 'ms-variants-per-sample-graph'
ID_INDEL_LENGTH_GRAPH = 'ms-indel-lengths-graph'
ID_ANNOTATIONS_GRAPH = 'ms-id-annotations-graph'
ID_SUBSTITUTIONS_GRAPH = 'ms-substitutions-graph'
ID_SLIDER_MIN_SAMPLES = 'ms-slider-min-samples-per-country'
ID_DROPDOWN_DATA_SOURCE = "ms-dropdown-data-source"
ID_DROPDOWN_COUNTRY = 'ms-dropdown-country'
ID_DROPDOWN_GENE = 'ms-dropdown-gene-overall-mutations'
ID_DROPDOWN_VARIANT_TYPE = 'ms-dropdown-variant-type'
ID_ACCUMULATED_SAMPLES_GRAPH = 'ms-accumulated-samples-per-country'
ID_DN_DS_GRAPH = 'ms-dn_ds_graph'
ID_APPLY_BUTTOM = 'ms-apply-buttom'

def get_tab_mutation_stats(queries: Queries, data_source: DataSource):
    return dbc.CardBody(
            children=[
                get_samples_tab_left_bar(queries, data_source),
                html.Div(
                    className="one columns",
                    children=[html.Br()]),
                get_mutation_stats_tab_graphs()
        ])


def get_mutation_stats_tab_graphs():
    return html.Div(
        className="nine columns",
        children=[
            html.Br(),
            html.Div(children=[
                html.Div(id=ID_VARIANTS_PER_SAMPLE_GRAPH, className="six columns"),
                html.Div(id=ID_SUBSTITUTIONS_GRAPH, className="five columns"),
            ]),
            html.Div(children=[
                html.Div(id=ID_INDEL_LENGTH_GRAPH, className="six columns"),
                html.Div(id=ID_ANNOTATIONS_GRAPH, className="five columns"),
            ]),
        ])


def get_samples_tab_left_bar(queries: Queries, data_source: DataSource):
    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            html.Div(
                dcc.Dropdown(
                    id=ID_DROPDOWN_DATA_SOURCE,
                    options=[{'label': data_source.name, 'value': data_source.name}],
                    value=data_source.name,
                    clearable=False,
                    multi=False,
                    disabled=True
                ), style={'display': 'none'}),
            dcc.Markdown("""Select one or more genes"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_GENE,
                options=[{'label': g.name, 'value': g.name} for g in queries.get_genes()],
                value=None,
                multi=True
            ),
            html.Br(),
            dcc.Markdown("""Select a variant type for mutations per sample and top mutations"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_VARIANT_TYPE,
                options=[
                    {'label': VariantType.SNV.name, 'value': VariantType.SNV.name},
                    {'label': VariantType.MNV.name, 'value': VariantType.MNV.name},
                    {'label': VariantType.INSERTION.name, 'value': VariantType.INSERTION.name},
                    {'label': VariantType.DELETION.name, 'value': VariantType.DELETION.name}
                ],
                value=None,
                multi=True
            ),
            html.Br(),
            html.Button('Apply', id=ID_APPLY_BUTTOM),
        ])


def set_callbacks_mutation_stats_tab(app, session: Session):

    queries = Queries(session=session)
    figures = MutationStatsFigures(queries)

    @app.callback(
        Output(ID_VARIANTS_PER_SAMPLE_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_GENE, 'value'),
            State(ID_DROPDOWN_VARIANT_TYPE, 'value')
        ],
        suppress_callback_exceptions=True
    )
    def update_variants_per_sample(_, data_source, genes, variant_types):
        return html.Div(children=figures.get_variants_per_sample_plot(
            data_source=data_source, genes=genes, variant_types=variant_types))

    @app.callback(
        Output(ID_SUBSTITUTIONS_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_GENE, 'value'),
            State(ID_DROPDOWN_VARIANT_TYPE, 'value')
        ],
        suppress_callback_exceptions=True
    )
    def update_substitutions(_, data_source, genes, variant_types):
        return html.Div(children=figures.get_substitutions_plot(
            data_source=data_source, genes=genes, variant_types=variant_types))

    @app.callback(
        Output(ID_INDEL_LENGTH_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_GENE, 'value')
        ],
        suppress_callback_exceptions=True
    )
    def update_indel_lengths(_, data_source, genes):
        return html.Div(children=figures.get_indels_lengths_plot(data_source=data_source, genes=genes))

    @app.callback(
        Output(ID_ANNOTATIONS_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_GENE, 'value')
        ],
        suppress_callback_exceptions=True
    )
    def update_annotation(_, data_source, genes):
        return html.Div(children=figures.get_annotations_plot(data_source=data_source, genes=genes))


import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from dash.dependencies import Output, Input, State
from sqlalchemy.orm import Session
from covigator.dashboard.figures.lineages import LineageFigures
from covigator.dashboard.tabs import get_mini_container, print_number
from covigator.database.model import DataSource
from covigator.database.queries import Queries

ID_APPLY_BUTTOM = 'lineages-apply-buttom'


ID_DROPDOWN_DATA_SOURCE = "lineages-dropdown-data-source"
ID_DROPDOWN_COUNTRY = 'lineages-dropdown-country'
ID_DROPDOWN_LINEAGE = 'lineages-dropdown-lineage'
ID_LINEAGES_GRAPH = 'lineages-graph'
ID_LINEAGES_TABLE = 'lineages-table'


def get_tab_lineages(queries: Queries, data_source: DataSource):
    return dbc.CardBody(
            children=[
                get_lineages_tab_left_bar(queries, data_source),
                html.Div(
                    className="one column",
                    children=[html.Br()]),
                get_lineages_tab_graphs()
        ])


def get_lineages_tab_graphs():
    return html.Div(
        className="nine columns",
        children=[
            html.Br(),
            html.Div(id=ID_LINEAGES_GRAPH),
            html.Br(),
            html.Div(id=ID_LINEAGES_TABLE),
        ])


def get_lineages_tab_left_bar(queries: Queries, data_source: DataSource):

    lineages = queries.get_lineages(source=data_source.name)

    return html.Div(
        className="two columns",
        children=[
            html.P("Lineage information is derived from the mutated sequence using Pangolin."),
            html.Br(),
            html.Div(
                html.Span(
                    children=[
                        get_mini_container(
                            title="Lineages",
                            value=print_number(len(lineages))
                        )
                        ])),
            html.Br(),
            html.Div(
                dcc.Dropdown(
                    id=ID_DROPDOWN_DATA_SOURCE,
                    options=[{'label': data_source.name, 'value': data_source.name}],
                    value=data_source.name,
                    clearable=False,
                    multi=False,
                    disabled=True
                ), style={'display': 'none'}),
            dcc.Markdown("""Select one or more countries"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_COUNTRY,
                options=[{'label': c, 'value': c} for c in queries.get_countries(data_source.name)],
                value=None,
                multi=True
            ),
            html.Br(),
            dcc.Markdown("""Select one or more lineages"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_LINEAGE,
                options=[{'label': c, 'value': c} for c in queries.get_lineages(data_source.name)],
                value=None,
                multi=True
            ),
            html.Br(),
            html.P("Select a single lineage to explore its corresponding mutations."),
            html.Button('Apply', id=ID_APPLY_BUTTOM),
        ])


def set_callbacks_lineages_tab(app, session: Session):

    queries = Queries(session=session)
    figures = LineageFigures(queries)

    countries_ena = queries.get_countries(DataSource.ENA.name)
    countries_gisaid = queries.get_countries(DataSource.GISAID.name)
    lineages_ena = queries.get_lineages(DataSource.ENA.name)
    lineages_gisaid = queries.get_lineages(DataSource.GISAID.name)

    @app.callback(
        Output(ID_DROPDOWN_COUNTRY, 'options'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'))
    def set_countries(source):
        """
        Updates the country drop down list when the data source is changed
        """
        countries = []
        if source == DataSource.ENA.name:
            countries = [{'label': c, 'value': c} for c in countries_ena]
        elif source == DataSource.GISAID.name:
            countries = [{'label': c, 'value': c} for c in countries_gisaid]
        return countries

    @app.callback(
        Output(ID_DROPDOWN_LINEAGE, 'options'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'))
    def set_lineages(source):
        """
        Updates the country drop down list when the data source is changed
        """
        lineages = []
        if source == DataSource.ENA.name:
            lineages = [{'label': c, 'value': c} for c in lineages_ena]
        elif source == DataSource.GISAID.name:
            lineages = [{'label': c, 'value': c} for c in lineages_gisaid]
        return lineages

    @app.callback(
        Output(ID_LINEAGES_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_COUNTRY, 'value'),
            State(ID_DROPDOWN_LINEAGE, 'value'),
        ],
        suppress_callback_exceptions=True
    )
    def update_lineages_plot(_, data_source, countries, lineages):
        return html.Div(children=figures.get_lineages_plot(
            data_source=data_source,
            countries=countries,
            lineages=lineages))

    @app.callback(
        Output(ID_LINEAGES_TABLE, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_COUNTRY, 'value'),
            State(ID_DROPDOWN_LINEAGE, 'value'),
        ],
        suppress_callback_exceptions=True
    )
    def update_lineages_table(_, data_source, countries, lineages):
        return html.Div(children=figures.get_lineages_variants_table(
            data_source=data_source, lineages=lineages, countries=countries))



import dash_bootstrap_components as dbc
import dash_html_components as html


def get_tab_overview():

    return dbc.CardBody([
            get_header(),
            dbc.Row([
                dbc.Col([None], width=2),
                dbc.Col([
                    html.Div(
                        children=[
                            html.Br(),
                            html.P(
                                """
                                Human infections with SARS-CoV-2 are spreading globally since the beginning of 2020, necessitating preventive or 
                                therapeutic strategies and first steps towards an end to this pandemic were done with the approval of the first mRNA 
                                vaccines against SARS-CoV-2. 
                                The accumulation of virus samples that have been sequenced in a short time frame is unprecedented.
                                This is the first pandemic recorded at a molecular level with such level of detail giving us the opportunity to develop
                                new tools for the monitoring of its evolution.
                                """),
                            html.P(
                                """
                                We want to provide an up-to-date interactive view on SARS-CoV-2 mutations to support global efforts in preventing or 
                                treating infections. 
                                Monitoring the appearance of relevant new mutations is key to enable a fast reaction to new strains and for that 
                                purpose we enable the exploration of these mutations and their annotations. 
                                Thus, we envision to help guiding global vaccine design efforts to overcome the threats of this pandemic.
                                """),
                            html.P(
                                """
                                CoVigator is a monitoring system for SARS-CoV-2 which integrates a full variant calling pipeline, 
                                a database that stores all relevant information about mutations in SARS-CoV-2 and finally a dashboard to enable 
                                visual analytics.
                                """),
                            html.Br(),
                            html.P("""
                       CoVigator loads publicly available SARS-CoV-2 DNA sequences from two systems: 
                       the European Nucleotide Archive (ENA) and the Global Initiative on Sharing Avian Influenza 
                       Data (GISAID). 
                       Some samples are present both in ENA and GISAID as some national initiatives are systematically
                       reporting to both databases. 
                       ENA provides the raw reads in FASTQ format and thus enables a higher resolution analysis into the 
                       SARS-CoV-2 mutations. Intrahost mutations are of particular interest. 
                       On the other hand, GISAID dataset is more extensive and represents best the geograhical and
                       temporal spread of SARS-CoV-2.
                       """),
                            html.Br(),
                            dbc.CardBody(
                                dbc.Row([
                                    dbc.Col(
                                        dbc.Card(
                                            [
                                                dbc.CardImg(src="/assets/gisaid.png", top=True,
                                                            style={"width": "18rem", "margin-left": "20px",
                                                                   "margin-top": "10px"}, ),
                                                dbc.CardBody(
                                                    [
                                                        dbc.Button(
                                                            "Explore data derived from GISAID", color="warning", href="/gisaid",
                                                            style={"margin-left": "20px", "margin-right": "20px",
                                                                   "font-size": 20}, ),
                                                    ]
                                                ),
                                            ],
                                            outline=False,
                                            style={"width": "40rem", "height": "15rem"},
                                        )
                                    ),
                                dbc.Col(
                                    dbc.Card(
                                        [
                                            dbc.CardImg(src="/assets/ENA_logo_2021.png", top=True,
                                                        style={"width": "23rem", "margin-left": "20px",
                                                               "margin-top": "10px"}, ),
                                            dbc.CardBody(
                                                [
                                                    dbc.Button(
                                                        "Explore data derived from ENA", color="warning", href="/ena",
                                                        style={"margin-left": "20px", "margin-right": "20px",
                                                               "font-size": 20}, ),
                                                ]
                                            ),
                                        ],
                                        outline=False,
                                        style={"width": "40rem", "height": "15rem"},
                                    )
                                )])),
                            html.Br(),
                            html.Br(),
                            html.Br(),
                            html.Br(),
                        ],
                        style={"text-align": "left", "font-size": 16}),
                ], width=8),
                dbc.Col([None], width=2),
            ], align='center'),
        ])


def get_header():
    return dbc.Row([
                dbc.Col([None], width=2),
                dbc.Col([html.Div(
                    [html.Div([html.H1("Monitoring SARS-CoV-2 mutations")], style={"text-align": "left"})],
                    id="title")], width=5),
                dbc.Col([None], width=2)
        ], id="header", className="row flex-display",)




class CovigatorException(Exception):
    pass


class CovigatorDatabaseConnectionException(CovigatorException):
    pass


class CovigatorPipelineError(CovigatorException):
    pass


class CovigatorDashBoardInitialisationError(CovigatorException):
    pass


class CovigatorNotSupportedVariant(CovigatorException):
    pass


class CovigatorQueryException(CovigatorException):
    pass


class CovigatorErrorProcessingCoverageResults(CovigatorException):
    pass


class CovigatorErrorProcessingPangolinResults(CovigatorException):
    pass


class CovigatorErrorProcessingDeduplicationResults(CovigatorException):
    pass


class CovigatorDashboardMissingPrecomputedData(CovigatorException):
    pass


class CovigatorExcludedSampleException(CovigatorException):
    pass


class CovigatorExcludedAssemblySequence(CovigatorExcludedSampleException):
    pass


class CovigatorExcludedSampleTooEarlyDateException(CovigatorExcludedSampleException):
    pass


class CovigatorExcludedSampleTooManyMutations(CovigatorExcludedSampleException):
    pass


class CovigatorExcludedSampleNarrowCoverage(CovigatorExcludedSampleException):
    pass


class CovigatorExcludedSampleBadQualityReads(CovigatorExcludedSampleException):
    pass



from covigator.database.queries import Queries
import plotly.graph_objects as go


PLOTLY_CONFIG = {
    'displaylogo': False,
    'displayModeBar': False,
    'modeBarButtonsToRemove': ['zoom', 'pan', 'select', 'zoomIn', 'zoomOut', 'autoScale', 'resetScale', 'lasso2d']
}
MARGIN = go.layout.Margin(l=0, r=0, b=0, t=30)
TEMPLATE = "plotly_white"

STYLES_STRIPPED = [{
                'if': {'row_index': 'odd'},
                'backgroundColor': 'rgb(248, 248, 248)'
            }]
STYLE_HEADER = {
    'backgroundColor': 'white',
    'fontWeight': 'bold'
}
STYLE_CELL = {
    'padding': '5px',
    'maxWidth': '100px'
}


class Figures:

    def __init__(self, queries: Queries):
        self.queries = queries



import abc
import time
import traceback
import pandas as pd
from contextlib import suppress
from datetime import datetime
from typing import Callable
import typing as typing
from dask.distributed import Client
from distributed import fire_and_forget
from logzero import logger
import covigator
import covigator.configuration
from covigator.configuration import Configuration
from covigator.database.database import Database, session_scope
from covigator.database.model import Log, DataSource, CovigatorModule, JobStatus, SampleGisaid, \
    SampleEna
from covigator.database.queries import Queries
from covigator.exceptions import CovigatorExcludedSampleException, CovigatorErrorProcessingPangolinResults
from covigator.precomputations.loader import PrecomputationsLoader


class AbstractProcessor:

    def __init__(self, database: Database, dask_client: Client, data_source: DataSource, config: Configuration,
                 download : bool = False, wait_time=60):
        self.data_source = data_source
        self.config = config
        self.start_time = datetime.now()
        self.has_error = False
        self.error_message = None
        self.database = database
        assert self.database is not None, "Empty database"
        self.dask_client = dask_client
        assert self.dask_client is not None, "Empty dask client"
        self.session = self.database.get_database_session()
        self.queries = Queries(self.session)
        self.wait_time = wait_time
        self.download = download

    def process(self):
        logger.info("Starting processor")
        count = 0
        count_batch = 0
        try:
            while True:
                # queries 100 jobs every time to make sending to queue faster
                jobs = self.queries.find_first_pending_jobs(
                    self.data_source, n=1000,
                    # only reads jobs in PENDING status if --download is indicated
                    status=[JobStatus.PENDING, JobStatus.DOWNLOADED] if self.download else [JobStatus.DOWNLOADED])
                if jobs is None or len(jobs) == 0:
                    logger.info("No more jobs to process after sending {} runs to process".format(count))
                    break
                for job in jobs:
                    # it has to update the status before doing anything so this processor does not read it again
                    job.status = JobStatus.QUEUED
                    job.queued_at = datetime.now()
                    self.session.commit()

                    # sends the run for processing
                    future = self._process_run(run_accession=job.run_accession)
                    fire_and_forget(future)
                    count += 1
                    count_batch += 1
                    if count % 1000 == 0:
                        logger.info("Sent {} jobs for processing...".format(count))

                    # waits for a batch to finish
                    if count_batch >= self.config.batch_size:
                        # waits for a batch to finish before sending more
                        self._wait_for_batch()
                        count_batch = 0

            # waits for the last batch to finish
            if count_batch > 0:
                self._wait_for_batch()
            logger.info("Processor finished!")

            # precomputes data right after processor
            PrecomputationsLoader(session=self.session).load()

        except Exception as e:
            logger.exception(e)
            self.session.rollback()
            self.error_message = self._get_traceback_from_exception(e)
            self.has_error = True
        finally:
            logger.info("Logging execution stats...")
            self._write_execution_log(count, data_source=self.data_source)
            logger.info("Waits {} secs to let the cluster tidy up things...".format(self.wait_time))
            time.sleep(self.wait_time)
            logger.info("Shutting down cluster and database session...")
            with suppress(Exception):
                self.dask_client.close(60)
                self.session.close()
            logger.info("Cluster and database sessions closed")

    def _wait_for_batch(self):
        logger.info("Waiting for a batch of jobs...")
        while (count_pending_jobs := self.queries.count_jobs_in_queue(data_source=self.data_source)) > 0:
            logger.info("Waiting for {} pending jobs".format(count_pending_jobs))
            time.sleep(self.wait_time)
        logger.info("Batch finished")

    @staticmethod
    def run_job(config: Configuration, run_accession: str, start_status: JobStatus, end_status: JobStatus,
                error_status: JobStatus, data_source: DataSource,
                function: Callable[[typing.Union[SampleEna, SampleGisaid] , Queries, Configuration],
                                   typing.Union[SampleEna, SampleGisaid]]) -> str or None:
        """
        Runs a function on a job, if anything goes wrong or does not fit in the DB it returns None in order to
        stop the execution of subsequent jobs.
        """
        covigator.configuration.initialise_logs(config.logfile_processor, sample_id=run_accession)
        if run_accession is not None:
            try:
                with session_scope(config=config) as session:
                    queries = Queries(session)
                    sample = queries.find_job_by_accession_and_status(
                        run_accession=run_accession, status=start_status, data_source=data_source)
                    if sample is not None:
                        sample = function(sample, queries, config)
                        if end_status is not None:
                            sample.status = end_status
                    else:
                        logger.warning("Expected ENA job {} in status {}".format(run_accession, start_status))
                        run_accession = None
            except CovigatorExcludedSampleException as e:
                # captures exclusion cases
                AbstractProcessor._log_error_in_job(
                    config=config, run_accession=run_accession, exception=e, status=JobStatus.EXCLUDED,
                    data_source=data_source)
                run_accession = None
            except Exception as e:
                # captures any possible exception happening, but logs it in the DB
                if error_status is not None:
                    AbstractProcessor._log_error_in_job(
                        config=config, run_accession=run_accession, exception=e, status=error_status,
                        data_source=data_source)
                    run_accession = None
                else:
                    logger.warning("Error processing a job that does not stop the workflow!")
        return run_accession

    @abc.abstractmethod
    def _process_run(self, run_accession: str):
        pass

    def _write_execution_log(self, count, data_source: DataSource):
        end_time = datetime.now()
        self.session.add(Log(
            start=self.start_time,
            end=end_time,
            source=data_source,
            module=CovigatorModule.PROCESSOR,
            has_error=self.has_error,
            processed=count,
            data={
                "processed": count
            }
        ))
        self.session.commit()

    @staticmethod
    def _get_traceback_from_exception(e):
        return "".join(traceback.format_exception(
            etype=type(e), value=e, tb=e.__traceback__))

    @staticmethod
    def _log_error_in_job(config: Configuration, run_accession: str, exception: Exception, status: JobStatus, data_source: DataSource):
        with session_scope(config=config) as session:
            logger.exception(exception)
            logger.info("Error on job {} on state {}: {}".format(run_accession, status, str(exception)))
            sample = Queries(session).find_job_by_accession(run_accession=run_accession, data_source=data_source)
            sample.status = status
            sample.failed_at = datetime.now()
            sample.error_message = AbstractProcessor._get_traceback_from_exception(exception)

    @staticmethod
    def load_pangolin(sample: typing.Union[SampleGisaid, SampleEna], path: str) -> typing.Union[SampleGisaid, SampleEna]:
        try:
            data = pd.read_csv(path,
                               na_values=None,
                               dtype={
                                   'lineage': str,
                                   'conflict': float,
                                   'ambiguity_score': float,
                                   'scorpio_call': str,
                                   'scorpio_support': float,
                                   'scorpio_conflict': float,
                                   'version': str,
                                   'pangolin_version': str,
                                   'pangoLEARN_version': str,
                                   'pango_version': str,
                                   'status': str,
                                   'note': str
                               })

            # fill NA values on a per column basis...
            data.lineage.fillna(value="", inplace=True)
            data.lineage = data.lineage.transform(lambda x: "" if x == 'None' else x)   # replace "None" values
            data.scorpio_call.fillna(value="", inplace=True)
            data.version.fillna(value="", inplace=True)
            data.pangolin_version.fillna(value="", inplace=True)
            data.pangoLEARN_version.fillna(value="", inplace=True)
            data.pango_version.fillna(value="", inplace=True)
            data.note.fillna(value="", inplace=True)
            data.conflict.fillna(value=0.0, inplace=True)
            data.ambiguity_score.fillna(value=0.0, inplace=True)
            data.scorpio_support.fillna(value=0.0, inplace=True)
            data.scorpio_conflict.fillna(value=0.0, inplace=True)

            sample.pangolin_lineage = data.lineage.loc[0]
            sample.pangolin_conflict = data.conflict.loc[0]
            sample.pangolin_ambiguity_score = data.ambiguity_score.loc[0]
            sample.pangolin_scorpio_call = data.scorpio_call.loc[0]
            sample.pangolin_scorpio_support = data.scorpio_support.loc[0]
            sample.pangolin_scorpio_conflict = data.scorpio_conflict.loc[0]
            sample.pangolin_version = data.version.loc[0]
            sample.pangolin_pangolin_version = data.pangolin_version.loc[0]
            sample.pangolin_pangoLEARN_version = data.pangoLEARN_version.loc[0]
            sample.pangolin_pango_version = data.pango_version.loc[0]
            sample.pangolin_status = data.status.loc[0]
            sample.pangolin_note = data.note.loc[0]
        except Exception as e:
            raise CovigatorErrorProcessingPangolinResults(e)

        return sample


from datetime import date, datetime
from typing import List, Union
import pandas as pd
from logzero import logger
import sqlalchemy
from sqlalchemy import and_, desc, asc, func, String, DateTime
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.orm import Session, aliased
from sqlalchemy.sql.sqltypes import NullType
from covigator import SYNONYMOUS_VARIANT
from covigator.database.model import DataSource, SampleEna, JobStatus, \
    VariantObservation, Gene, Variant, VariantCooccurrence, Conservation, SampleGisaid, \
    SubclonalVariantObservation, PrecomputedVariantsPerSample, PrecomputedSubstitutionsCounts, PrecomputedIndelLength, \
    VariantType, PrecomputedAnnotation, PrecomputedOccurrence, PrecomputedTableCounts, \
    PrecomputedVariantAbundanceHistogram, PrecomputedSynonymousNonSynonymousCounts, RegionType, Domain, \
    GisaidVariantObservation, GisaidVariant, LastUpdate, GisaidVariantCooccurrence
from covigator.exceptions import CovigatorQueryException, CovigatorDashboardMissingPrecomputedData


class Queries:

    def __init__(self, session: Session):
        self.session = session

    @staticmethod
    def get_variant_observation_klass(source: str):
        if source == DataSource.ENA.name:
            klass = VariantObservation
        elif source == DataSource.GISAID.name:
            klass = GisaidVariantObservation
        else:
            raise CovigatorQueryException("Bad data source: {}".format(source))
        return klass

    @staticmethod
    def get_variant_klass(source: str):
        if source == DataSource.ENA.name:
            klass = Variant
        elif source == DataSource.GISAID.name:
            klass = GisaidVariant
        else:
            raise CovigatorQueryException("Bad data source: {}".format(source))
        return klass

    @staticmethod
    def get_sample_klass(source: str):
        if source == DataSource.ENA.name:
            klass = SampleEna
        elif source == DataSource.GISAID.name:
            klass = SampleGisaid
        else:
            raise CovigatorQueryException("Bad data source: {}".format(source))
        return klass

    @staticmethod
    def get_variant_cooccurrence_klass(source: str):
        if source == DataSource.ENA.name:
            klass = VariantCooccurrence
        elif source == DataSource.GISAID.name:
            klass = GisaidVariantCooccurrence
        else:
            raise CovigatorQueryException("Bad data source: {}".format(source))
        return klass

    def find_job_by_accession_and_status(
            self, run_accession: str, status: JobStatus, data_source: DataSource) -> Union[SampleEna, SampleGisaid]:
        klass = self.get_sample_klass(source=data_source.name)
        return self.session.query(klass)\
            .filter(and_(klass.run_accession == run_accession, klass.status == status)).first()

    def find_job_by_accession(self, run_accession: str, data_source: DataSource) -> Union[SampleEna, SampleGisaid]:
        klass = self.get_sample_klass(source=data_source.name)
        return self.session.query(klass).filter(klass.run_accession == run_accession).first()

    def find_first_by_status(self, data_source: DataSource, status, n=100) -> List[Union[SampleEna, SampleGisaid]]:
        klass = self.get_sample_klass(source=data_source.name)
        return self.session.query(klass) \
            .filter(klass.status.in_(status)) \
            .order_by(klass.created_at.desc()) \
            .limit(n) \
            .all()

    def find_first_pending_jobs(
            self, data_source: DataSource, n=100, status: List = [JobStatus.DOWNLOADED]) -> List[Union[SampleEna, SampleGisaid]]:
        return self.find_first_by_status(data_source=data_source, status=status, n=n)

    def find_first_jobs_to_download(self, data_source: DataSource, n=100) -> List[Union[SampleEna, SampleGisaid]]:
        return self.find_first_by_status(data_source=data_source, status=[JobStatus.PENDING], n=n)

    def count_jobs_in_queue(self, data_source):
        return self.count_jobs_by_status(data_source=data_source, status=JobStatus.QUEUED)

    def count_jobs_by_status(self, data_source: DataSource, status: JobStatus):
        klass = self.get_sample_klass(source=data_source.name)
        return self.session.query(klass).filter(klass.status == status).count()

    def find_sample_by_accession(self, run_accession: str, source: DataSource) -> Union[SampleEna, SampleGisaid]:
        klass = self.get_sample_klass(source=source.name)
        return self.session.query(klass).filter(klass.run_accession == run_accession).first()

    def get_countries(self, source: str) -> List[str]:
        klass = self.get_sample_klass(source=source)
        return [c for c, in self.session.query(klass.country).filter(
                klass.status == JobStatus.FINISHED.name).distinct().order_by(klass.country.asc()).all()]

    def get_lineages(self, source: str) -> List[str]:
        klass = self.get_sample_klass(source=source)
        return [c for c, in self.session.query(klass.pangolin_lineage).filter(
            and_(klass.status == JobStatus.FINISHED.name, klass.pangolin_lineage != None)).distinct().order_by(
                klass.pangolin_lineage.asc()).all()]

    def get_variants_per_sample(self, data_source: str, genes: List[str], variant_types: List[str]):
        """
        Returns a DataFrame with columns: number_mutations, count, type
        where type: SNV, insertion or deletion
        """
        self._assert_data_source(data_source)
        query = self.session.query(PrecomputedVariantsPerSample)\
            .filter(PrecomputedVariantsPerSample.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedVariantsPerSample.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedVariantsPerSample.gene_name == None)
        if variant_types is not None and variant_types:
            query = query.filter(PrecomputedVariantsPerSample.variant_type.in_(variant_types))

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            data.variant_type = data.variant_type.transform(lambda x: x.name if x else x)
        return data[["number_mutations", "variant_type", "count"]] \
            .groupby(["number_mutations", "variant_type"]).sum().reset_index()

    def _assert_data_source(self, data_source):
        if data_source != DataSource.ENA.name and data_source != DataSource.GISAID.name:
            raise CovigatorQueryException("Bad data source:  {}".format(data_source))

    def get_indel_lengths(self, data_source, genes):
        self._assert_data_source(data_source)
        query = self.session.query(PrecomputedIndelLength).filter(PrecomputedIndelLength.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedIndelLength.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedIndelLength.gene_name == None)

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            data["tmp_variant_type"] = data["length"].transform(
                lambda x: VariantType.INSERTION.name if x > 0 else VariantType.DELETION.name)
            data["inframe"] = data["length"].transform(lambda x: "INFRAME" if x % 3 == 0 else "FRAMESHIFT")
            data["variant_type"] = data[["tmp_variant_type", "inframe"]].apply(
                lambda x: "{}_{}".format(x[0], x[1]), axis=1)
            return data[["length", "variant_type", "count"]] \
                .groupby(["length", "variant_type", ]).sum().reset_index() \
                .sort_values("length", ascending=True)
        return data

    def get_annotations(self, data_source, genes):
        self._assert_data_source(data_source)
        query = self.session.query(PrecomputedAnnotation).filter(PrecomputedAnnotation.source == data_source)
        if genes is not None and genes:
            query = query.filter(PrecomputedAnnotation.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedAnnotation.gene_name == None)

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            return data[["annotation", "count"]] \
                .groupby(["annotation", ]).sum().reset_index() \
                .sort_values("count", ascending=False)
        return data

    def get_substitutions(self, data_source, genes, variant_types):
        self._assert_data_source(data_source)
        query = self.session.query(PrecomputedSubstitutionsCounts)\
            .filter(PrecomputedSubstitutionsCounts.source == data_source)
        if variant_types is not None and variant_types:
            query = query.filter(PrecomputedSubstitutionsCounts.variant_type.in_(variant_types))
        if genes is not None and genes:
            query = query.filter(PrecomputedSubstitutionsCounts.gene_name.in_(genes))
        else:
            query = query.filter(PrecomputedSubstitutionsCounts.gene_name == None)

        data = pd.read_sql(query.statement, self.session.bind)
        if data.shape[0] > 0:
            data.variant_type = data.variant_type.transform(lambda x: x.name if x else x)
            data["substitution"] = data[["reference", "alternate"]].apply(lambda x: "{}>{}".format(x[0], x[1]), axis=1)
            total_number_variants = data["count"].sum()
            data = data[["substitution", "variant_type", "count"]] \
                .groupby(["substitution", "variant_type"]).sum().reset_index()\
                .sort_values("count", ascending=False)
            data["rate"] = (data["count"] / total_number_variants).transform(lambda x: "{} %".format(round(x * 100, 1)))
            data = data.head(12)
        return data

    def get_accumulated_samples_by_country(
            self, data_source: str, countries: List[str], lineages: List[str], min_samples=100) -> pd.DataFrame:
        """
        Returns a DataFrame with columns: data, country, cumsum, count
        """
        klass = self.get_sample_klass(source=data_source)
        query = self.session.query(
            func.count().label("count"), klass.collection_date.label("date"), klass.country) \
            .filter(klass.status == JobStatus.FINISHED.name) \
            .group_by(klass.collection_date, klass.country)
        if countries:
            query = query.filter(klass.country.in_(countries))
        if lineages:
            query = query.filter(klass.pangolin_lineage.in_(lineages))
        samples = pd.read_sql(query.statement, self.session.bind).astype({'date': 'datetime64', 'count': 'float64'})

        filled_table = None
        if samples is not None and samples.shape[0] > 0:
            # accumulates count ordered by date
            samples['cumsum'] = samples.groupby(['country'])['count'].cumsum()

            # creates empty table with all pairwise combinations of date and country
            dates = samples.date[~samples.date.isna()].sort_values().unique()
            countries = samples[(~samples.country.isna()) & (samples["cumsum"] >= min_samples)] \
                .sort_values("cumsum", ascending=False).country.unique()

            empty_table = pd.DataFrame(
                index=pd.MultiIndex.from_product([dates, countries], names=["date", "country"]))
            empty_table["count"] = 0

            # adds values into empty table
            filled_table = pd.merge(
                left=empty_table.reset_index(),
                right=samples.loc[:, ["date", "country", "count"]],
                on=["date", "country"],
                how='left',
                suffixes=("_x", "_y")).fillna(0)

            filled_table["count"] = filled_table.count_x + filled_table.count_y
            filled_table['cumsum'] = filled_table.groupby(['country'])['count'].cumsum()

        return filled_table

    def get_accumulated_lineages_by_country(
            self, data_source: str, countries: List[str], lineages: List[str], min_samples=100) -> pd.DataFrame:
        """
        Returns a DataFrame with columns: data, country, cumsum, count
        """
        klass = self.get_sample_klass(source=data_source)
        query = self.session.query(
            func.count().label("count"), klass.collection_date.label("date"), klass.pangolin_lineage.label("lineage")) \
            .filter(and_(klass.status == JobStatus.FINISHED.name, klass.pangolin_lineage != None)) \
            .group_by(klass.collection_date, klass.pangolin_lineage)
        if countries:
            query = query.filter(klass.country.in_(countries))
        if lineages:
            query = query.filter(klass.pangolin_lineage.in_(lineages))
        samples = pd.read_sql(query.statement, self.session.bind).astype({'date': 'datetime64', 'count': 'float64'})

        filled_table = None
        if samples is not None and samples.shape[0] > 0:
            # accumulates count ordered by date
            samples['cumsum'] = samples.groupby(['lineage'])['count'].cumsum()

            # creates empty table with all pairwise combinations of date and country
            dates = samples.date[~samples.date.isna()].sort_values().unique()

            lineages = samples[(~samples.lineage.isna()) & (samples["cumsum"] >= min_samples)] \
                .sort_values("cumsum", ascending=False).lineage.unique()

            empty_table = pd.DataFrame(
                index=pd.MultiIndex.from_product([dates, lineages], names=["date", "lineage"]))
            empty_table["count"] = 0

            # adds values into empty table
            filled_table = pd.merge(
                left=empty_table.reset_index(),
                right=samples.loc[:, ["date", "lineage", "count"]],
                on=["date", "lineage"],
                how='left',
                suffixes=("_x", "_y")).fillna(0)

            filled_table["count"] = filled_table.count_x + filled_table.count_y
            filled_table['cumsum'] = filled_table.groupby(['lineage'])['count'].cumsum()

            # add total samples per day
            counts_per_date = filled_table[["date", "count"]].groupby("date").sum().reset_index()
            counts_per_date.rename(columns={'count': 'total_per_date'}, inplace=True)
            filled_table = pd.merge(
                left=filled_table,
                right=counts_per_date,
                on="date")
            filled_table['ratio_per_date'] = filled_table[["count", "total_per_date"]].apply(
                lambda x: float(x[0]) / x[1], axis=1)

        return filled_table

    def get_sample_months(self, pattern, data_source: str) -> List[datetime]:
        klass = self.get_sample_klass(source=data_source)
        dates = [d.strftime(pattern) for d, in self.session.query(klass.collection_date).filter(
            and_(klass.status == JobStatus.FINISHED.name, klass.collection_date.isnot(None))).distinct().all()]
        return sorted(list(set(dates)))

    def get_gene(self, gene_name: str) -> Gene:
        return self.session.query(Gene).filter(Gene.name == gene_name).first()

    
    def get_genes(self) -> List[Gene]:
        return self.session.query(Gene).order_by(Gene.start).all()

    
    def get_genes_df(self) -> pd.DataFrame:
        return pd.read_sql(self.session.query(Gene).order_by(Gene.start).statement, self.session.bind)

    
    def get_domains_df(self) -> pd.DataFrame:
        return pd.read_sql(self.session.query(Domain).order_by(Domain.start).statement, self.session.bind)

    
    def get_domain(self, domain_name: str) -> Domain:
        return self.session.query(Domain).filter(Domain.name == domain_name).first()

    
    def get_domains(self) -> List[Domain]:
        return self.session.query(Domain).order_by(Domain.gene_name, Domain.start).all()

    
    def get_domains_by_gene(self, gene_name: str) -> List[Domain]:
        return self.session.query(Domain).filter(Domain.gene_name == gene_name).order_by(Domain.start).all()

    
    def get_non_synonymous_variants_by_region(self, start, end, source) -> pd.DataFrame:

        klass = self.get_variant_observation_klass(source)
        query = self.session.query(
            klass.position, klass.annotation_highest_impact, klass.hgvs_p, func.count().label("count_occurrences"))\
            .filter(and_(klass.annotation_highest_impact != SYNONYMOUS_VARIANT,
                         klass.position >= start, klass.position <= end))

        subquery = query.group_by(klass.position, klass.annotation_highest_impact, klass.hgvs_p).subquery()
        return pd.read_sql(
            self.session.query(subquery).filter(subquery.c.count_occurrences > 1).statement, self.session.bind)

    def get_variant_ids_by_sample(self, sample_id, source: str, maximum_length: int) -> List[str]:
        klass = self.get_variant_observation_klass(source=source)
        return self.session.query(klass.variant_id) \
            .filter(and_(klass.sample == sample_id, klass.length < maximum_length, klass.length > -maximum_length)) \
            .order_by(klass.position, klass.reference, klass.alternate) \
            .all()

    def increment_variant_cooccurrence(
            self, variant_id_one: str, variant_id_two: str, source: str):

        klazz = self.get_variant_cooccurrence_klass(source=source)
        variant_cooccurrence = self.session.query(klazz) \
            .filter(and_(klazz.variant_id_one == variant_id_one,
                         klazz.variant_id_two == variant_id_two)) \
            .first()
        if variant_cooccurrence is None:
            variant_cooccurrence = klazz(
                variant_id_one=variant_id_one,
                variant_id_two=variant_id_two,
                count=1)
            self.session.add(variant_cooccurrence)
        else:
            # NOTE: it is important to increase the counter like this to avoid race conditions
            # the increase happens in the database server and not in python
            # see https://stackoverflow.com/questions/2334824/how-to-increase-a-counter-in-sqlalchemy
            variant_cooccurrence.count = klazz.count + 1
        self.session.commit()
    
    def count_samples(self, source: str, cache=True) -> int:
        self._assert_data_source(source)
        if cache:
            query = self.session.query(PrecomputedTableCounts.count)
            if source == DataSource.ENA.name:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == SampleEna.__name__,
                    PrecomputedTableCounts.factor == PrecomputedTableCounts.FACTOR_SOURCE
                ))
            elif source == DataSource.GISAID.name:
                query = query.filter(and_(
                    PrecomputedTableCounts.table == SampleGisaid.__name__,
                    PrecomputedTableCounts.factor == PrecomputedTableCounts.FACTOR_SOURCE
                ))
            result = query.first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            klass = self.get_sample_klass(source=source)
            count = self.session.query(klass).filter(klass.status == JobStatus.FINISHED.name).count()
        return count

    def count_countries(self, source: str = None, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count).filter(and_(
                    PrecomputedTableCounts.table == PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY,
                    PrecomputedTableCounts.factor == PrecomputedTableCounts.FACTOR_SOURCE,
                    PrecomputedTableCounts.value == source
                ))
            result = query.first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            count = len(self.get_countries(source=source))
        return count

    def count_variants(self, source: str, cache=True):
        klass = self.get_variant_klass(source=source)
        if cache:
            result = self.session.query(PrecomputedTableCounts.count) \
                .filter(and_(
                PrecomputedTableCounts.table == klass.__name__, PrecomputedTableCounts.value == source)).first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            count = self.session.query(klass).count()
        return count

    def count_variant_observations(self, source: str = None, cache=True):
        klass = self.get_variant_observation_klass(source)
        if cache:
            result = self.session.query(PrecomputedTableCounts.count).filter(
                PrecomputedTableCounts.table == klass.__name__).first()
            if result is None:
                raise CovigatorDashboardMissingPrecomputedData
            count = result.count
        else:
            count = self.session.query(klass).count()
        return count

    def count_subclonal_variant_observations(self, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count) \
                .filter(PrecomputedTableCounts.table == SubclonalVariantObservation.__name__)
            result = query.first()
            count = result.count
        else:
            count = self.session.query(SubclonalVariantObservation).count()
        return count

    def count_unique_subclonal_variant(self, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count) \
                .filter(PrecomputedTableCounts.table == SubclonalVariantObservation.__name__ + "_unique")
            result = query.first()
            count = result.count
        else:
            count = self.session.query(SubclonalVariantObservation)\
                .distinct(SubclonalVariantObservation.variant_id).count()
        return count

    def count_unique_only_subclonal_variant(self, cache=True):
        if cache:
            query = self.session.query(PrecomputedTableCounts.count) \
                .filter(PrecomputedTableCounts.table == SubclonalVariantObservation.__name__ + "_unique_only_subclonal")
            result = query.first()
            count = result.count
        else:
            sql_query = """
                    SELECT COUNT(*) as count from (
                        select distinct variant_id 
                        from {subclonal_variants_table}
                        where variant_id not in (
                            select distinct variant_id from {variant_observations_table}
                        )
                        and vaf >= 0.03
                    ) as variants""".format(
                subclonal_variants_table=SubclonalVariantObservation.__tablename__,
                variant_observations_table=VariantObservation.__tablename__)
            count = int(pd.read_sql_query(sql_query, self.session.bind)["count"][0])
        return count

    def get_date_of_first_sample(self, source: DataSource = DataSource.ENA) -> date:
        """
        Returns the date of the earliest ENA sample loaded in the database
        """
        klass = self.get_sample_klass(source=source.name)
        result = self.session.query(klass.collection_date).filter(
            and_(klass.status == JobStatus.FINISHED.name, klass.collection_date.isnot(None))).order_by(asc(klass.collection_date)).first()
        return result[0] if result is not None else result

    def get_date_of_most_recent_sample(self, source: DataSource = DataSource.ENA) -> date:
        """
        Returns the date of the latest ENA sample loaded in the database
        """
        klass = self.get_sample_klass(source=source.name)
        result = self.session.query(klass.collection_date).filter(
            and_(klass.status == JobStatus.FINISHED.name, klass.collection_date.isnot(None))).order_by(desc(klass.collection_date)).first()
        return result[0] if result is not None else result

    def get_last_update(self, data_source: DataSource) -> date:
        result = self.session.query(LastUpdate.update_time).filter(LastUpdate.source == data_source).order_by(
            desc(LastUpdate.update_time)).first()
        most_recent_update = result[0] if result is not None else result
        return most_recent_update

    def get_variant_counts_by_month(self, variant_id, source: str) -> pd.DataFrame:

        klass = self.get_variant_observation_klass(source=source)
        sql_query_ds_ena = """
        select count(*) as count, variant_id, date_trunc('month', date::timestamp) as month 
            from {variant_observation_table} 
            where variant_id='{variant_id}'
            group by variant_id, date_trunc('month', date::timestamp);
            """.format(
            variant_observation_table=klass.__tablename__,
            variant_id=variant_id
        )
        data = pd.read_sql_query(sql_query_ds_ena, self.session.bind)
        data['month'] = pd.to_datetime(data['month'], utc=True)
        return data[~data.month.isna()]

    def get_sample_counts_by_month(self, source: str) -> pd.DataFrame:
        klass = self.get_sample_klass(source=source)
        query = self.session.query(
            func.date_trunc('month', klass.collection_date).label("month"),
            func.count().label("sample_count"))\
            .filter(klass.status == JobStatus.FINISHED.name) \
            .group_by(func.date_trunc('month', klass.collection_date))
        counts = pd.read_sql(query.statement, self.session.bind)
        counts['month'] = pd.to_datetime(counts['month'], utc=True)
        return counts

    def get_sparse_cooccurrence_matrix(self, gene_name, domain, min_cooccurrence=5) -> pd.DataFrame:
        """
        Returns the sparse cooccurrence matrix of all non synonymous variants in a gene or domain with at least
        min_occurrences occurrences.
        """
        # query for cooccurrence matrix
        variant_one = aliased(Variant)
        variant_two = aliased(Variant)
        query = self.session.query(VariantCooccurrence,
                                   variant_one.position,
                                   variant_one.reference,
                                   variant_one.alternate,
                                   variant_one.hgvs_p,
                                   variant_one.hgvs_p.label("hgvs_p_one"),
                                   variant_two.hgvs_p.label("hgvs_p_two"),
                                   (variant_one.hgvs_p + " - " + variant_two.hgvs_p).label("hgvs_tooltip"))

        query = query.filter(VariantCooccurrence.count >= min_cooccurrence) \
            .join(variant_one, and_(VariantCooccurrence.variant_id_one == variant_one.variant_id)) \
            .join(variant_two, and_(VariantCooccurrence.variant_id_two == variant_two.variant_id))

        if domain is not None:
            query = query.filter(and_(
                variant_one.pfam_name == domain,
                variant_one.annotation != SYNONYMOUS_VARIANT,
                variant_two.pfam_name == domain,
                variant_two.annotation != SYNONYMOUS_VARIANT))
        elif gene_name is not None:
            query = query.filter(and_(
                variant_one.gene_name == gene_name,
                variant_one.annotation != SYNONYMOUS_VARIANT,
                variant_two.gene_name == gene_name,
                variant_two.annotation != SYNONYMOUS_VARIANT))
        else:
            query = query.filter(and_(
                variant_one.annotation != SYNONYMOUS_VARIANT,
                variant_two.annotation != SYNONYMOUS_VARIANT))

        self._print_query(query=query)
        data = pd.read_sql(query.statement, self.session.bind)

        return data

    def get_top_occurring_variants_precomputed(
            self, top=10, gene_name=None, domain=None, metric="count", source=None) -> pd.DataFrame:
        """
        Returns the top occurring variants + the segregated counts of occurrences per month
        with columns: chromosome, position, reference, alternate, total, month, count
        """
        query = self.session.query(PrecomputedOccurrence).filter(PrecomputedOccurrence.source == source)
        # if domain is provided it supersedes the gene filter
        if domain is not None:
            query = query.filter(PrecomputedOccurrence.domain == domain)
        elif gene_name is not None:
            query = query.filter(PrecomputedOccurrence.gene_name == gene_name)
        if metric == "count":
            query = query.order_by(PrecomputedOccurrence.count.desc())
        elif metric == "frequency_by_month":
            query = query.order_by(PrecomputedOccurrence.frequency.desc())
        else:
            raise CovigatorQueryException("Not supported metric for top occurring variants")

        top_occurring_variants = pd.read_sql(query.statement, self.session.bind)

        # formats the DNA mutation
        top_occurring_variants.rename(columns={'variant_id': 'dna_mutation'}, inplace=True)
        top_occurring_variants["frequency_by_month"] = top_occurring_variants.frequency

        # pivots the table over months
        top_occurring_variants = pd.pivot_table(
            top_occurring_variants, index=['gene_name', 'dna_mutation', 'hgvs_p', 'annotation', "frequency", "total"],
            columns=["month"], values=[metric], fill_value=0).droplevel(0, axis=1).reset_index()

        return top_occurring_variants.sort_values(by="frequency", ascending=False).head(top)

    def get_variant_abundance_histogram(self, source: str, bin_size=50, cache=True) -> pd.DataFrame:
        histogram = None
        if cache:
            query = self.session.query(PrecomputedVariantAbundanceHistogram)
            if source is not None:
                query = query.filter(and_(PrecomputedVariantAbundanceHistogram.bin_size == bin_size,
                                          PrecomputedVariantAbundanceHistogram.source == source))
            else:
                query = query.filter(and_(PrecomputedVariantAbundanceHistogram.bin_size == bin_size,
                                          PrecomputedVariantAbundanceHistogram.source == None))
            histogram = pd.read_sql(query.statement, self.session.bind)
            if histogram.shape[0] == 0:
                raise CovigatorDashboardMissingPrecomputedData
            histogram = histogram[["position_bin", "count_unique_variants", "count_variant_observations"]]
        else:
            klass = self.get_variant_klass(source=source)
            klass_observation = self.get_variant_observation_klass(source=source)

            # queries for the maximum position
            maximum_position = self.session.query(func.max(klass.position)).first()[0]
            if maximum_position is not None:
                # builds all possible bins
                all_bins = pd.DataFrame(
                    data=[i*bin_size for i in range(int(maximum_position/bin_size) + 1)],
                    columns=["position_bin"])

                # counts variants over those bins
                sql_query = """
                        SELECT cast("position"/{bin_size} as int)*{bin_size} AS position_bin,
                               COUNT(*) as count_unique_variants
                        FROM {table_name}
                        GROUP BY position_bin
                        ORDER BY position_bin;
                        """.format(bin_size=bin_size, table_name=klass.__tablename__)
                binned_counts_variants = pd.read_sql_query(sql_query, self.session.bind)

                # counts variant observations over those bins
                sql_query = """
                        SELECT cast("position"/{bin_size} as int)*{bin_size} AS position_bin,
                               COUNT(*) as count_variant_observations
                        FROM {table_name}
                        GROUP BY position_bin
                        ORDER BY position_bin;
                        """.format(bin_size=bin_size, table_name=klass_observation.__tablename__)
                binned_counts_variant_observations = pd.read_sql_query(sql_query, self.session.bind)

                histogram = pd.merge(
                    left=pd.merge(
                        left=all_bins,
                        right=binned_counts_variants,
                        on="position_bin", how="left").reset_index(),
                    right=binned_counts_variant_observations,
                    on="position_bin", how="left").reset_index()
                histogram.fillna(0, inplace=True)
                histogram = histogram[["position_bin", "count_unique_variants", "count_variant_observations"]]

        return histogram

    def get_conservation_table(self, bin_size=50, start=None, end=None) -> pd.DataFrame:
        # counts variants over those bins
        sql_query = """
                SELECT cast("start"/{bin_size} as int)*{bin_size} AS position_bin,
                       AVG("conservation") as conservation,
                       AVG("conservation_sarbecovirus") as conservation_sarbecovirus,
                       AVG("conservation_vertebrates") as conservation_vertebrates
                FROM {table_name}
                {where}
                GROUP BY position_bin
                ORDER BY position_bin;
                """.format(bin_size=bin_size, table_name= Conservation.__tablename__,
                           where="WHERE start >= {start} and start <= {end}".format(start=start, end=end)
                           if start is not None and end is not None else "")
        return pd.read_sql_query(sql_query, self.session.bind)

    def get_dnds_table(self, source: str, countries=None, genes=None) -> pd.DataFrame:
        self._assert_data_source(data_source=source)
        # counts variants over those bins
        query_genes = self.session.query(PrecomputedSynonymousNonSynonymousCounts)\
            .filter(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.GENE)
        query_domains = self.session.query(PrecomputedSynonymousNonSynonymousCounts) \
            .filter(PrecomputedSynonymousNonSynonymousCounts.region_type == RegionType.DOMAIN)
        data_domains = None

        if source is not None:
            query_genes = query_genes.filter(PrecomputedSynonymousNonSynonymousCounts.source == source)
            query_domains = query_domains.filter(PrecomputedSynonymousNonSynonymousCounts.source == source)
        if countries is not None and len(countries) > 0:
            query_genes = query_genes.filter(PrecomputedSynonymousNonSynonymousCounts.country.in_(countries))
            query_domains = query_domains.filter(PrecomputedSynonymousNonSynonymousCounts.country.in_(countries))
        if genes is not None and len(genes) > 0:
            query_genes = query_genes.filter(PrecomputedSynonymousNonSynonymousCounts.region_name.in_(genes))
            domains = []
            for g in genes:
                domains = domains + [d.name for d in self.get_domains_by_gene(gene_name=g)]
            query_domains = query_domains.filter(PrecomputedSynonymousNonSynonymousCounts.region_name.in_(domains))
            data_domains = pd.read_sql(query_domains.statement, self.session.bind)

        data = pd.read_sql(query_genes.statement, self.session.bind)
        if data_domains is not None and data_domains.shape[1] > 0:
            data = pd.concat([data, data_domains])

        return data

    def _print_query(self, query):
        class StringLiteral(String):
            """Teach SA how to literalize various things."""

            def literal_processor(self, dialect):
                super_processor = super(StringLiteral, self).literal_processor(dialect)

                def process(value):
                    if isinstance(value, int):
                        return str(value)
                    if not isinstance(value, str):
                        value = str(value)
                    result = super_processor(value)
                    if isinstance(result, bytes):
                        result = result.decode(dialect.encoding)
                    return result

                return process

        class LiteralDialect(DefaultDialect):
            colspecs = {
                # prevent various encoding explosions
                String: StringLiteral,
                # teach SA about how to literalize a datetime
                DateTime: StringLiteral,
                # don't format py2 long integers to NULL
                NullType: StringLiteral,
            }
        logger.info(query.statement.compile(
            dialect=LiteralDialect(),
            compile_kwargs={'literal_binds': True}).string)

    def column_windows(self, session, column, windowsize):
        """Return a series of WHERE clauses against
        a given column that break it into windows.

        Result is an iterable of tuples, consisting of
        ((start, end), whereclause), where (start, end) are the ids.

        Requires a database that supports window functions,
        i.e. Postgresql, SQL Server, Oracle.

        Enhance this yourself !  Add a "where" argument
        so that windows of just a subset of rows can
        be computed.

        """
        def int_for_range(start_id, end_id):
            if end_id:
                return and_(
                    column >= start_id,
                    column < end_id
                )
            else:
                return column >= start_id

        q = session.query(
            column,
            func.row_number(). \
                over(order_by=column). \
                label('rownum')
        ). \
            from_self(column)
        if windowsize > 1:
            q = q.filter(sqlalchemy.text("rownum %% %d=1" % windowsize))

        intervals = [id for id, in q]

        while intervals:
            start = intervals.pop(0)
            if intervals:
                end = intervals[0]
            else:
                end = None
            yield int_for_range(start, end)

    def windowed_query(self, query, column, windowsize):
        """"
        Break a Query into windows on a given column.

        This magic comes from here: https://github.com/sqlalchemy/sqlalchemy/wiki/RangeQuery-and-WindowedRangeQuery
        """
        for whereclause in self.column_windows(query.session, column, windowsize):
            for row in query.filter(whereclause).order_by(column):
                yield row



import pandas as pd
from logzero import logger
from sqlalchemy import func, and_
from sqlalchemy.orm import Session
from covigator import SYNONYMOUS_VARIANT
from covigator.database.model import DataSource, PrecomputedVariantsPerLineage
from covigator.database.queries import Queries


class VariantsPerLineageLoader:

    def __init__(self, session: Session):
        self.session = session
        self.queries = Queries(session=self.session)

    def load(self):

        database_rows = self._read_variants_per_lineage()

        # delete all rows before starting
        self.session.query(PrecomputedVariantsPerLineage).delete()
        self.session.commit()

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedVariantsPerLineage.__tablename__))

    def _read_variants_per_lineage(self):
        # gets the top occurrent variants for each source
        variants_per_lineage_ena = None
        try:
            variants_per_lineage_ena = self.get_variants_per_lineage(source=DataSource.ENA.name)
        except ValueError as e:
            logger.exception(e)
            logger.error("No top occurrences for ENA data")
        variants_per_lineage_gisaid = None
        try:
            variants_per_lineage_gisaid = self.get_variants_per_lineage(source=DataSource.GISAID.name)
        except ValueError:
            logger.error("No top occurrences for GISAID data")
        database_rows = []
        # stores the precomputed data
        if variants_per_lineage_ena is not None:
            for index, row in variants_per_lineage_ena.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_variants_per_lineage(row, source=DataSource.ENA))
        if variants_per_lineage_gisaid is not None:
            for index, row in variants_per_lineage_gisaid.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_variants_per_lineage(row, source=DataSource.GISAID))
        return database_rows

    def _row_to_variants_per_lineage(self, row, source):
        return PrecomputedVariantsPerLineage(
            lineage=row["lineage"],
            variant_id=row["variant_id"],
            country=row["country"],
            count_observations=row["count_observations"],
            source=source
        )

    def get_variants_per_lineage(self, source: str):
        klass_variant_observation = self.queries.get_variant_observation_klass(source)
        klass_sample = self.queries.get_sample_klass(source)
        query = self.session.query(
            klass_variant_observation.variant_id,
            klass_sample.country,
            klass_sample.pangolin_lineage.label("lineage"),
            func.count().label('count_observations'))\
            .filter(and_(
                klass_variant_observation.annotation_highest_impact != SYNONYMOUS_VARIANT,
                klass_sample.pangolin_lineage != None,
                klass_sample.pangolin_lineage != "")) \
            .join(klass_sample, klass_variant_observation.sample == klass_sample.run_accession) \
            .group_by(klass_sample.pangolin_lineage,
                      klass_sample.country,
                      klass_variant_observation.variant_id)
        variants_per_lineage = pd.read_sql(query.statement, self.session.bind)
        return variants_per_lineage



import logging

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import logzero
from sqlalchemy.orm import Session
from dash.dependencies import Input, Output
import covigator
import covigator.configuration
from covigator.configuration import Configuration
from covigator.dashboard.tabs.acknowledgements import get_tab_acknowledgements
from covigator.dashboard.tabs.dataset_ena import get_tab_dataset_ena
from covigator.dashboard.tabs.dataset_gisaid import get_tab_dataset_gisaid
from covigator.dashboard.tabs.download import set_callbacks_download_tab, get_tab_download
from covigator.dashboard.tabs.footer import get_footer
from covigator.dashboard.tabs.lineages import set_callbacks_lineages_tab, get_tab_lineages
from covigator.dashboard.tabs.mutation_stats import get_tab_mutation_stats, set_callbacks_mutation_stats_tab
from covigator.dashboard.tabs.overview import get_tab_overview
from covigator.dashboard.tabs.samples import get_tab_samples, set_callbacks_samples_tab
from covigator.dashboard.tabs.intrahost_mutations import get_tab_subclonal_variants, set_callbacks_subclonal_variants_tab
from covigator.dashboard.tabs.recurrent_mutations import get_tab_variants, set_callbacks_variants_tab
from covigator.database.database import get_database
from logzero import logger

from covigator.database.model import DataSource
from covigator.database.queries import Queries

TAB_STYLE = {"color": "#003c78", 'margin-right': '15px'}

ID_TAB_CONTENT = "tab-content"
DOWNLOAD_TAB_ID = "download"
HELP_TAB_ID = "help"
INTRAHOST_MUTATIONS_TAB_ID = "subclonal-variants"
RECURRENT_MUTATIONS_TAB_ID = "variants"
SAMPLES_TAB_ID = "samples"
LINEAGES_TAB_ID = "lineages"
MUTATIONS_TAB_ID = "mutation-stats"
GISAID_DATASET_TAB_ID = "gisaid-dataset"
ENA_DATASET_TAB_ID = "ena-dataset"
OVERVIEW_TAB_ID = "overview"


class Dashboard:

    def __init__(self, config: Configuration, verbose=False):
        covigator.configuration.initialise_logs(config.logfile_dash)
        self.config = config
        # the connection to the database is created only once, but multiple sessions are used
        self.database = get_database(config=config, initialize=True, verbose=verbose)

    def serve_layout(self):
        logger.info("Serving layout")
        footer = get_footer()

        layout = html.Div(children=[
            dbc.Card([
                dbc.CardHeader(
                    children=[
                        dcc.Location(id='url', refresh=False),
                        dbc.Navbar(
                            [
                                dbc.Row([
                                        dbc.Col(
                                            children=html.A(html.Img(src="/assets/CoVigator_logo_txt_reg_no_bg.png",
                                                                     height="25px"), href="/"),
                                            className="ml-2",
                                            id="logo"
                                        ),
                                        #dbc.Col(html.Br(), className="ml-2"),
                                        #dbc.Col(html.Br(), className="ml-2"),
                                    ],
                                    align="left",
                                    className="g-0",
                                    style={'align': 'left'}
                                ),
                                dbc.Row(
                                    [
                                        dbc.Tabs(
                                            None,
                                            id="tabs",
                                            active_tab=SAMPLES_TAB_ID,
                                            style={'align': 'right', 'font-size': '140%'},
                                        )],
                                    align="right",
                                    style={'margin-left': '2%', }
                                ),
                                dbc.Row([
                                    dbc.Col(
                                        None,
                                        className="ml-2",
                                        id="top-right-logo",
                                        align="left",
                                        style={'margin-left': '2%', }
                                    ),
                                    dbc.Col(html.Br()),
                                    dbc.Col(
                                        dbc.DropdownMenu(
                                            label="Menu", children=[
                                                dbc.DropdownMenuItem(
                                                    "Home", href="/", class_name="m-1",
                                                    style={'font-size' : '150%', "color": "#003c78"}),
                                                dbc.DropdownMenuItem(
                                                    "GISAID dataset", href="/gisaid",
                                                    style={'font-size' : '150%', "color": "#003c78"}),
                                                dbc.DropdownMenuItem(
                                                    "ENA dataset", href="/ena",
                                                    style={'font-size' : '150%', "color": "#003c78"}),
                                                dbc.DropdownMenuItem(
                                                    "Documentation", href="https://covigator.readthedocs.io/en/latest",
                                                    target="_blank",
                                                    style={'font-size': '150%', "color": "#003c78"}),
                                                dbc.DropdownMenuItem(
                                                    "Acknowledgements", href="/acknowledgements",
                                                    style={'font-size' : '150%', "color": "#003c78"}),
                                            ],
                                            align_end=True,
                                            size="lg",
                                            toggle_style={
                                                "textTransform": "uppercase",
                                                "background": "#003c78",
                                                'font-size': '85%'
                                            },
                                            style={"margin-left": "15px"}
                                        ),
                                        style={}
                                    )],
                                    align="right",
                                    justify="end",
                                    style={'float': 'right', 'position': 'absolute', 'right': 0, 'text-align': 'right',
                                           },
                                    className="g-0 ms-auto flex-nowrap mt-3 mt-md-0",
                                )
                                ]
                        )],
                ),
                dbc.CardBody(dcc.Loading(id="loading-1", children=[html.Div(id=ID_TAB_CONTENT)], style={"height": "100%"})),
                dbc.CardFooter(footer)
            ])
        ])
        return layout

    def start_dashboard(self, debug=False):
        try:
            if debug:
                logzero.loglevel(level=logging.DEBUG)
            logger.info("Starting covigator dashboard")
            app = self.get_application()
            app.run_server(debug=debug, host=self.config.dash_host, port=self.config.dash_port)
        except Exception as e:
            logger.exception(e)
            raise e

    def get_application(self) -> dash.Dash:

        # creates the Dash application
        logger.info("Create the application")
        app = dash.Dash(
            name=__name__,
            title="CoVigator",
            external_stylesheets=[dbc.themes.BOOTSTRAP],
            suppress_callback_exceptions=True,
            meta_tags=[
                # A description of the app, used by e.g.
                # search engines when displaying search results.
                {
                    'name': 'description',
                    'content': 'CoVigator - monitoring SARS-CoV-2 mutations'
                },
                # A tag that tells Internet Explorer (IE)
                # to use the latest renderer version available
                # to that browser (e.g. Edge)
                {
                    'http-equiv': 'X-UA-Compatible',
                    'content': 'IE=edge'
                },
                # A tag that tells the browser not to scale
                # desktop widths to fit mobile screens.
                # Sets the width of the viewport (browser)
                # to the width of the device, and the zoom level
                # (initial scale) to 1.
                #
                # Necessary for "true" mobile support.
                {
                  'name': 'viewport',
                  'content': 'width=device-width, initial-scale=1.0'
                }
            ]
        )
        # Warning pass the layout as a function, do not call it otherwise the application will serve a static dataset
        app.layout = self.serve_layout
        session = self.database.get_database_session()
        set_callbacks(app=app, session=session, content_folder=self.config.content_folder)
        set_callbacks_variants_tab(app=app, session=session)
        set_callbacks_samples_tab(app=app, session=session)
        set_callbacks_lineages_tab(app=app, session=session)
        set_callbacks_mutation_stats_tab(app=app, session=session)
        set_callbacks_subclonal_variants_tab(app=app, session=session)
        set_callbacks_download_tab(app=app, content_folder=self.config.content_folder)
        return app


def set_callbacks(app, session: Session, content_folder):

    queries = Queries(session=session)

    MAIN_PAGE = "main"
    ENA_PAGE = DataSource.ENA
    GISAID_PAGE = DataSource.GISAID
    ACKNOWLEDGEMENTS_PAGE = "acknowledgements"

    def _get_page(url):
        if url in ["", "/"]:
            return MAIN_PAGE
        elif url == "/gisaid":
            return GISAID_PAGE
        elif url == "/ena":
            return ENA_PAGE
        elif url == "/acknowledgements":
            return ACKNOWLEDGEMENTS_PAGE
        else:
            raise ValueError("This URL does not exist")

    @app.callback(
        Output('tabs', "children"),
        Output('tabs', "active_tab"),
        [Input("url", "pathname")])
    def switch_page(url):
        page = _get_page(url)
        if page == MAIN_PAGE:
            # show overview with links
            return None, None
        elif page == GISAID_PAGE:
            # show gisaid tabs
            return [
                dbc.Tab(label="Overview", tab_id=GISAID_DATASET_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Samples", tab_id=SAMPLES_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Lineages", tab_id=LINEAGES_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Mutation statistics", tab_id=MUTATIONS_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Recurrent mutations", tab_id=RECURRENT_MUTATIONS_TAB_ID, label_style=TAB_STYLE),
                ], GISAID_DATASET_TAB_ID
        elif page == ENA_PAGE:
            # show ena tabs
            return [
               dbc.Tab(label="Overview", tab_id=ENA_DATASET_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Samples", tab_id=SAMPLES_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Lineages", tab_id=LINEAGES_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Mutation statistics", tab_id=MUTATIONS_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Recurrent mutations", tab_id=RECURRENT_MUTATIONS_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Intrahost mutations", tab_id=INTRAHOST_MUTATIONS_TAB_ID, label_style=TAB_STYLE),
                dbc.Tab(label="Download data", tab_id=DOWNLOAD_TAB_ID, label_style=TAB_STYLE)], ENA_DATASET_TAB_ID
        elif page == ACKNOWLEDGEMENTS_PAGE:
            # show ena tabs
            return [
                dbc.Tab(label="Acknowledgements", tab_id=HELP_TAB_ID, label_style={"color": "#003c78", 'display': 'none'})], HELP_TAB_ID

    @app.callback(
        Output('logo', "children"),
        [Input("url", "pathname")])
    def switch_logo(url):
        page = _get_page(url)
        if page == MAIN_PAGE or page == ACKNOWLEDGEMENTS_PAGE:
            return html.A(html.Img(src="/assets/CoVigator_logo_txt_reg_no_bg.png", height="80px"), href="/")
        elif page == GISAID_PAGE:
            return html.A(html.Img(src="/assets/CoVigator_logo_GISAID.png", height="80px"), href="/")
        elif page == ENA_PAGE:
            return html.A(html.Img(src="/assets/CoVigator_logo_ENA.png", height="80px"), href="/")

    @app.callback(
        Output('top-right-logo', "children"),
        [Input("url", "pathname")])
    def switch_lat_update(url):
        page = _get_page(url)
        if page == MAIN_PAGE:
            return None
        elif page == GISAID_PAGE:
            return dbc.Button(
                "last updated {date}".format(date=queries.get_last_update(DataSource.GISAID)),
                outline=True, color="dark", className="me-1",
                # 'background-color': '#b71300',
                style={"margin-right": "15px", 'font-size': '85%'})
        elif page == ENA_PAGE:
            return dbc.Button(
                "last updated {date}".format(date=queries.get_last_update(DataSource.ENA)),
                outline=True, color="dark", className="me-1",
                style={"margin-right": "15px", 'font-size': '85%'})

    @app.callback(
        Output(ID_TAB_CONTENT, "children"),
        [Input("tabs", "active_tab"), Input("url", "pathname")])
    def switch_tab(at, url):
        page = _get_page(url)
        try:
            if page == MAIN_PAGE:
                return get_tab_overview()
            elif at == ENA_DATASET_TAB_ID:
                return get_tab_dataset_ena(queries=queries)
            elif at == GISAID_DATASET_TAB_ID:
                return get_tab_dataset_gisaid(queries=queries)
            elif at == SAMPLES_TAB_ID:
                return get_tab_samples(queries=queries, data_source=page)
            elif at == MUTATIONS_TAB_ID:
                return get_tab_mutation_stats(queries=queries, data_source=page)
            elif at == RECURRENT_MUTATIONS_TAB_ID:
                return get_tab_variants(queries=queries, data_source=page)
            elif at == INTRAHOST_MUTATIONS_TAB_ID:
                return get_tab_subclonal_variants(queries=queries)
            elif at == DOWNLOAD_TAB_ID:
                return get_tab_download(content_folder=content_folder)
            elif at == HELP_TAB_ID:
                return get_tab_acknowledgements()
            elif at == LINEAGES_TAB_ID:
                return get_tab_lineages(queries=queries, data_source=page)
            return html.P("This shouldn't ever be displayed...")
        except Exception as e:
            logger.exception(e)


def main(debug=False):
    Dashboard(config=Configuration()).start_dashboard(debug=debug)


if __name__ == '__main__':
    main(debug=True)



from covigator.database.model import VariantType

VARIANT_TYPE_COLOR_MAP = {
    VariantType.SNV.name: "#8da0cb",
    VariantType.INSERTION.name: "#fc8d62",
    VariantType.DELETION.name: "#66c2a5",
}


from datetime import date
from covigator.database.model import SampleEna, Log, DataSource, CovigatorModule
from covigator.tests.unit_tests.abstract_test import AbstractTest
from covigator.tests.unit_tests.faked_objects import FakeEnaAccessor


class EnaAccessorTests(AbstractTest):

    def test_filtering_by_library_strategies(self):
        ena_accessor = FakeEnaAccessor([
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "AMPLICON",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "OTHER",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/004/ERR4080484/ERR4080484_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ])
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 1)
        self.assertEqual(ena_accessor.excluded, 2)
        self.assertEqual(ena_accessor.excluded_samples_by_library_strategy.get("AMPLICON"), 1)
        self.assertEqual(ena_accessor.excluded_samples_by_library_strategy.get("OTHER"), 1)

    def test_filtering_by_instrument_platform(self):
        ena_accessor = FakeEnaAccessor([
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "OXFORD_NANOPORE",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "SANGER",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/004/ERR4080484/ERR4080484_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ])
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 1)
        self.assertEqual(ena_accessor.excluded, 2)
        self.assertEqual(ena_accessor.excluded_samples_by_instrument_platform.get("OXFORD_NANOPORE"), 1)
        self.assertEqual(ena_accessor.excluded_samples_by_instrument_platform.get("SANGER"), 1)

    def test_filtering_by_host_taxid(self):
        ena_accessor = FakeEnaAccessor([
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/004/ERR4080484/ERR4080484_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "1111"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "2222"}
        ])
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 1)
        self.assertEqual(ena_accessor.excluded, 2)
        self.assertEqual(ena_accessor.excluded_samples_by_host_tax_id.get("1111"), 1)
        self.assertEqual(ena_accessor.excluded_samples_by_host_tax_id.get("2222"), 1)

    def test_filtering_by_missing_fastqs(self):
        ena_accessor = FakeEnaAccessor([
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": None,
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ])
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 1)
        self.assertEqual(ena_accessor.excluded, 2)
        self.assertEqual(ena_accessor.excluded_samples_by_fastq_ftp, 2)

    def test_no_filtering(self):
        ena_accessor = FakeEnaAccessor([
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ])
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 3)
        self.assertEqual(ena_accessor.excluded, 0)

    def test_filtering_data_already_in_db(self):
        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ], database=self.database)
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 3)
        self.assertEqual(ena_accessor.excluded, 0)

        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "THIS_IS_A_NEW_ONE",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ], database=self.database)
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 1)
        self.assertEqual(ena_accessor.excluded_existing, 2)

    def test_country_parsing(self):
        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606",
             "country": "england"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606",
             "country": "GermaN"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606",
             "country": "Morocco:Meknez"},
            {"run_accession": "ERR4080486",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080487",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606",
             "country": ""},
            {"run_accession": "ERR4080488",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606",
             "country": "Jupiter"}
        ], database=self.database)
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 6)
        self.assertEqual(ena_accessor.excluded, 0)
        session = self.database.get_database_session()
        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080483").first()
        self.assertEqual(run.country_raw, "england")
        self.assertEqual(run.country, "United Kingdom")
        self.assertEqual(run.country_alpha_2, "GB")
        self.assertEqual(run.country_alpha_3, "GBR")
        self.assertEqual(run.continent_alpha_2, "EU")
        self.assertEqual(run.continent, "Europe")
        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080484").first()
        self.assertEqual(run.country_raw, "GermaN")
        self.assertEqual(run.country, "Germany")
        self.assertEqual(run.country_alpha_2, "DE")
        self.assertEqual(run.country_alpha_3, "DEU")
        self.assertEqual(run.continent_alpha_2, "EU")
        self.assertEqual(run.continent, "Europe")
        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080485").first()
        self.assertEqual(run.country_raw, "Morocco:Meknez")
        self.assertEqual(run.country, "Morocco")
        self.assertEqual(run.country_alpha_2, "MA")
        self.assertEqual(run.country_alpha_3, "MAR")
        self.assertEqual(run.continent_alpha_2, "AF")
        self.assertEqual(run.continent, "Africa")
        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080486").first()
        self.assertEqual(run.country_raw, None)
        self.assertEqual(run.country, "Not available")
        self.assertEqual(run.country_alpha_2, "None")
        self.assertEqual(run.country_alpha_3, "None")
        self.assertEqual(run.continent_alpha_2, "None")
        self.assertEqual(run.continent, "None")
        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080487").first()
        self.assertEqual(run.country_raw, "")
        self.assertEqual(run.country, "Not available")
        self.assertEqual(run.country_alpha_2, "None")
        self.assertEqual(run.country_alpha_3, "None")
        self.assertEqual(run.continent_alpha_2, "None")
        self.assertEqual(run.continent, "None")
        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080488").first()
        self.assertEqual(run.country_raw, "Jupiter")
        self.assertEqual(run.country, "Not available")
        self.assertEqual(run.country_alpha_2, "None")
        self.assertEqual(run.country_alpha_3, "None")
        self.assertEqual(run.continent_alpha_2, "None")
        self.assertEqual(run.continent, "None")

    def test_dates_parsing(self):
        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606",
             "first_created": "2020-01-01",
             "collection_date": "2019-12-31",
             },
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606",
             "first_created": "2020-01-01 14:50",
             "collection_date": "2019-12-31 12:12:12"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606",
             "first_created": "blah",
             "collection_date": "blah"}
        ], database=self.database)
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 3)
        self.assertEqual(ena_accessor.excluded, 0)
        session = self.database.get_database_session()

        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080483").first()
        self.assertEqual(run.collection_date, date.fromisoformat("2019-12-31"))
        self.assertEqual(run.first_created, date.fromisoformat("2020-01-01"))

        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080484").first()
        self.assertIsNone(run.collection_date)
        self.assertIsNone(run.first_created)

        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080485").first()
        self.assertIsNone(run.collection_date)
        self.assertIsNone(run.first_created)

    def test_numeric_values(self):
        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606",
             "lat": "",
             "lon": "",
             "read_count": "",
             "base_count": "",
             "nominal_length": "",
             },
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606",
             "lat": "hey",
             "lon": "hey",
             "read_count": "hey",
             "base_count": "hey",
             "nominal_length": "hey"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606",
             "lat": "1.1",
             "lon": "1.1",
             "read_count": "1",
             "base_count": "1",
             "nominal_length": "1"
             }
        ], database=self.database)
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 3)
        self.assertEqual(ena_accessor.excluded, 0)
        session = self.database.get_database_session()

        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080483").first()
        self.assertIsNone(run.lat)
        self.assertIsNone(run.lon)
        self.assertIsNone(run.nominal_length)
        self.assertIsNone(run.read_count)
        self.assertIsNone(run.base_count)

        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080484").first()
        self.assertIsNone(run.lat)
        self.assertIsNone(run.lon)
        self.assertIsNone(run.nominal_length)
        self.assertIsNone(run.read_count)
        self.assertIsNone(run.base_count)

        run = session.query(SampleEna).filter(SampleEna.run_accession == "ERR4080485").first()
        self.assertEqual(run.lat, 1.1)
        self.assertEqual(run.lon, 1.1)
        self.assertEqual(run.nominal_length, 1)
        self.assertEqual(run.read_count, 1)
        self.assertEqual(run.base_count, 1)

    def test_sample_and_job_loading(self):
        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"
             },
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"
            },
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"
             }
        ], database=self.database)
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 3)
        self.assertEqual(ena_accessor.excluded, 0)
        session = self.database.get_database_session()

        self._assert_entities_from_accessor(session, "ERR4080483")
        self._assert_entities_from_accessor(session, "ERR4080484")
        self._assert_entities_from_accessor(session, "ERR4080485")

    def _assert_entities_from_accessor(self, session, identifier):
        self.assertEqual(session.query(SampleEna).filter(SampleEna.run_accession == identifier).count(), 1)

    def test_writing_logs(self):
        ena_accessor = FakeEnaAccessor(results=[
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"
             },
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080483/ERR4080483_1.fastq.gz",
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"
            },
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"
             }
        ], database=self.database)
        ena_accessor.access()

        session = self.database.get_database_session()
        self.assertEqual(session.query(Log).count(), 1)
        log = session.query(Log).first()
        self.assertIsNotNone(log.start)
        self.assertIsNotNone(log.end)
        self.assertEqual(log.source, DataSource.ENA)
        self.assertEqual(log.module, CovigatorModule.ACCESSOR)
        self.assertEqual(log.processed, 3)
        data = log.data
        self.assertEqual(data.get("included"), 3)
        self.assertEqual(data.get("excluded").get("existing"), 0)
        self.assertEqual(data.get("excluded").get("excluded_by_criteria"), 0)
        self.assertEqual(data.get("excluded").get("missing_fastq"), 0)
        self.assertIsInstance(data.get("excluded").get("platform"), dict)
        self.assertIsInstance(data.get("excluded").get("library_strategy"), dict)
        self.assertIsInstance(data.get("excluded").get("host"), dict)

    def test_excluding_rnaseq_samples(self):
        ena_accessor = FakeEnaAccessor([
            {"run_accession": "ERR4080483",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "ILLUMINA",
             "library_strategy": "RNA-Seq",
             "fastq_ftp": "",
             "fastq_md5": "a91a9dfa2f7008e13a7ce9767aa9aaf3",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080484",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "RNA-Seq",
             "library_strategy": "WGS",
             "fastq_ftp": None,
             "fastq_md5": "c57fef34933cbbec2e9e08867f3c664c",
             "host_tax_id": "9606"},
            {"run_accession": "ERR4080485",
             "scientific_name": "Severe acute respiratory syndrome coronavirus 2",
             "instrument_platform": "RNA-Seq",
             "library_strategy": "WGS",
             "fastq_ftp": "ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/005/ERR4080485/ERR4080485_1.fastq.gz",
             "fastq_md5": "4de269d2b5831e1c5175586af694d21e",
             "host_tax_id": "9606"}
        ])
        ena_accessor.access()
        self.assertEqual(ena_accessor.included, 0)
        self.assertEqual(ena_accessor.excluded, 3)



from unittest import TestCase
import pkg_resources
import covigator.tests
from covigator.database.database import Database
from covigator.database.model import Variant, VariantObservation, DataSource
from covigator.pipeline.vcf_loader import VcfLoader


class VcfLoaderTests(TestCase):

    def test_vcf_loader_all_variants(self):
        session = Database(test=True).get_database_session()
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/Sars_cov_2.all_variants.snpeff.vcf.gz")
        VcfLoader().load(vcf_file, run_accession="TEST1", source=DataSource.ENA, session=session)
        session.commit()
        self.assertEqual(session.query(Variant).count(), 89709)
        self.assertEqual(session.query(VariantObservation).count(), 89709)



import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from dash.dependencies import Output, Input, State
from sqlalchemy.orm import Session
from covigator.dashboard.figures.samples import SampleFigures
from covigator.database.model import DataSource
from covigator.database.queries import Queries

ID_APPLY_BUTTOM = 's-apply-buttom'


ID_SLIDER_MIN_SAMPLES = 'slider-min-samples-per-country'
ID_DROPDOWN_DATA_SOURCE = "dropdown-data-source"
ID_DROPDOWN_COUNTRY = 'dropdown-country'
ID_DROPDOWN_LINEAGE = 'dropdown-lineage'
ID_DROPDOWN_GENE = 'dropdown-gene-overall-mutations'
ID_ACCUMULATED_SAMPLES_GRAPH = 'accumulated-samples-per-country'
ID_DN_DS_GRAPH = 'dn_ds_graph'


def get_tab_samples(queries: Queries, data_source: DataSource):
    return dbc.CardBody(
            children=[
                get_samples_tab_left_bar(queries, data_source),
                html.Div(
                    className="one column",
                    children=[html.Br()]),
                get_samples_tab_graphs()
        ])


def get_samples_tab_graphs():
    return html.Div(
        className="nine columns",
        children=[
            html.Br(),
            html.Div(id=ID_ACCUMULATED_SAMPLES_GRAPH),
            html.Br(),
            html.Div(id=ID_DN_DS_GRAPH),
        ])


def get_samples_tab_left_bar(queries: Queries, data_source: DataSource):
    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            html.Div(
                dcc.Dropdown(
                    id=ID_DROPDOWN_DATA_SOURCE,
                    options=[{'label': data_source.name, 'value': data_source.name}],
                    value=data_source.name,
                    clearable=False,
                    multi=False,
                    disabled=True
                ), style={'display': 'none'}),
            dcc.Markdown("""Select one or more countries"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_COUNTRY,
                options=[{'label': c, 'value': c} for c in queries.get_countries(data_source.name)],
                value=None,
                multi=True
            ),
            html.Br(),
            dcc.Markdown("""Select one or more lineages"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_LINEAGE,
                options=[{'label': c, 'value': c} for c in queries.get_lineages(data_source.name)],
                value=None,
                multi=True
            ),
            html.Br(),
            dcc.Markdown("""Minimum number of samples per country"""),
            dcc.Slider(
                id=ID_SLIDER_MIN_SAMPLES,
                min=0,
                max=10000,
                step=100,
                value=100,
                dots=False,
                marks={i: '{}'.format(i) for i in [0, 2500, 5000, 7500, 10000]},
                tooltip=dict(always_visible=True, placement="right")
            ),
            html.Br(),
            dcc.Markdown("""Select one or more genes to show the dN/dS ratio on protein domains"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_GENE,
                options=[{'label': g2, 'value': g2} for g2 in sorted([g.name for g in queries.get_genes()])],
                value=None,
                multi=True
            ),
            html.Br(),
            html.Button('Apply', id=ID_APPLY_BUTTOM),
        ])


def set_callbacks_samples_tab(app, session: Session):

    queries = Queries(session=session)
    figures = SampleFigures(queries)

    countries_ena = queries.get_countries(DataSource.ENA.name)
    countries_gisaid = queries.get_countries(DataSource.GISAID.name)
    lineages_ena = queries.get_lineages(DataSource.ENA.name)
    lineages_gisaid = queries.get_lineages(DataSource.GISAID.name)

    @app.callback(
        Output(ID_DROPDOWN_COUNTRY, 'options'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'))
    def set_countries(source):
        """
        Updates the country drop down list when the data source is changed
        """
        countries = []
        if source == DataSource.ENA.name:
            countries = [{'label': c, 'value': c} for c in countries_ena]
        elif source == DataSource.GISAID.name:
            countries = [{'label': c, 'value': c} for c in countries_gisaid]
        return countries

    @app.callback(
        Output(ID_DROPDOWN_LINEAGE, 'options'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'))
    def set_lineages(source):
        """
        Updates the country drop down list when the data source is changed
        """
        lineages = []
        if source == DataSource.ENA.name:
            lineages = [{'label': c, 'value': c} for c in lineages_ena]
        elif source == DataSource.GISAID.name:
            lineages = [{'label': c, 'value': c} for c in lineages_gisaid]
        return lineages

    @app.callback(
        Output(ID_SLIDER_MIN_SAMPLES, 'disabled'),
        Input(ID_DROPDOWN_COUNTRY, 'value'),
        Input(ID_DROPDOWN_LINEAGE, 'value')
    )
    def disable_minimum_number_samples(countries, lineages):
        """
        Disables the minimum number of samples option when countries are provided
        """
        return not _is_empty_list(countries) or not _is_empty_list(lineages)

    def _is_empty_list(my_list):
        return my_list is None or len(my_list) == 0

    @app.callback(
        Output(ID_ACCUMULATED_SAMPLES_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_COUNTRY, 'value'),
            State(ID_SLIDER_MIN_SAMPLES, 'value'),
            State(ID_DROPDOWN_LINEAGE, 'value'),
        ],
        suppress_callback_exceptions=True
    )
    def update_accumulated_samples_by_country(_, data_source, countries, min_samples, lineages):
        return html.Div(children=figures.get_accumulated_samples_by_country_plot(
            data_source=data_source,
            countries=countries,
            min_samples=min_samples if _is_empty_list(countries) and _is_empty_list(lineages) else 0,
            lineages=lineages))

    @app.callback(
        Output(ID_DN_DS_GRAPH, 'children'),
        inputs=[Input(ID_APPLY_BUTTOM, 'n_clicks')],
        state=[
            State(ID_DROPDOWN_DATA_SOURCE, 'value'),
            State(ID_DROPDOWN_COUNTRY, 'value'),
            State(ID_DROPDOWN_GENE, 'value'),
        ],
        suppress_callback_exceptions=True
    )
    def update_dn_ds_graph(_, data_source, countries, genes):
        return html.Div(children=figures.get_dnds_by_gene_plot(
            data_source=data_source, countries=countries, genes=genes))



import pandas as pd
from datetime import datetime

import covigator
from covigator.configuration import Configuration
from covigator.database.model import JobStatus, DataSource, SampleGisaid
from covigator.database.database import Database
from logzero import logger
from dask.distributed import Client

from covigator.database.queries import Queries
from covigator.exceptions import CovigatorErrorProcessingPangolinResults
from covigator.processor.abstract_processor import AbstractProcessor
from covigator.pipeline.gisaid_pipeline import GisaidPipeline
from covigator.pipeline.vcf_loader import VcfLoader


class GisaidProcessor(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration, wait_time=60):
        logger.info("Initialising GISAID processor")
        super().__init__(database, dask_client, DataSource.GISAID, config=config, wait_time=wait_time)

    def _process_run(self, run_accession: str):
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(GisaidProcessor.job, self.config, run_accession, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession):
        return GisaidProcessor.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=DataSource.GISAID,
            function=GisaidProcessor.run_all)

    @staticmethod
    def run_all(sample: SampleGisaid, queries: Queries, config: Configuration):
        sample = GisaidProcessor.run_pipeline(sample=sample, queries=queries, config=config)
        sample = GisaidProcessor.load(sample=sample, queries=queries, config=config)
        return sample

    @staticmethod
    def run_pipeline(sample: SampleGisaid, queries: Queries, config: Configuration) -> SampleGisaid:
        pipeline_results = GisaidPipeline(config=config).run(sample=sample)
        sample.analysed_at = datetime.now()
        sample.sample_folder = sample.get_sample_folder(config.storage_folder)
        sample.vcf_path = pipeline_results.vcf_path
        sample.pangolin_path = pipeline_results.pangolin_path
        sample.fasta_path = pipeline_results.fasta_path

        # stores the covigator version
        sample.covigator_processor_version = covigator.VERSION

        # load pangolin results
        sample = GisaidProcessor.load_pangolin(sample=sample, path=sample.pangolin_path)

        return sample

    @staticmethod
    def load(sample: SampleGisaid, queries: Queries, config: Configuration) -> SampleGisaid:
        if not config.skip_vcf_loading:
            VcfLoader().load(
                vcf_file=sample.vcf_path, run_accession=sample.run_accession, source=DataSource.GISAID, session=queries.session)
            sample.loaded_at = datetime.now()
        return sample



import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from covigator.dashboard.figures import VARIANT_TYPE_COLOR_MAP
from covigator.dashboard.figures.figures import PLOTLY_CONFIG, MARGIN, TEMPLATE
from covigator.dashboard.tabs import get_mini_container, print_number, print_date
from covigator.database.model import DataSource, SAMPLE_ENA_TABLE_NAME, VariantObservation, SampleEna, JobStatus
from covigator.database.queries import Queries
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go


LIBRARY_STRATEGY_COLOR_MAP = {
    "WGA": "#377eb8",
    "WGS": "#4daf4a",
    "RNA-Seq": "#984ea3",
    "Targeted-Capture": "#ff7f00",
}

LIBRARY_LAYOUT_COLOR_MAP = {
    "PAIRED": "#969696",
    "SINGLE": "#cccccc"
}


def get_tab_dataset_ena(queries: Queries):

    return dbc.CardBody(
        children=[
            get_ena_overview_tab_left_bar(queries),
            html.Div(
                className="one column",
                children=[html.Br()]),
            get_ena_overview_tab_graphs(queries=queries)
        ])


def get_ena_overview_tab_graphs(queries):
    return html.Div(
        className="nine columns",
        children=get_dataset_ena_tab_graphs(queries))


def get_ena_overview_tab_left_bar(queries: Queries):
    count_samples = queries.count_samples(source=DataSource.ENA.name)
    count_variants = queries.count_variants(source=DataSource.ENA.name)
    count_variant_observations = queries.count_variant_observations(source=DataSource.ENA.name)
    date_of_first_sample = queries.get_date_of_first_sample(source=DataSource.ENA)
    date_of_most_recent_sample = queries.get_date_of_most_recent_sample(source=DataSource.ENA)

    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            dcc.Markdown("""
                        The ENA database provides the raw reads and metadata from SARS-CoV-2 samples through the ENA 
                        Portal API (https://www.ebi.ac.uk/ena/portal/api/). 
                        FASTQ files containing the raw reads were downloaded and MD5 checked.
                        The processing pipeline runs reads trimming (fastp), alignment (BWA mem), 
                        coverage analysis (samtools), variant calling (LoFreq), normalization (vt and BCFtools),
                        annotation (SnpEff and VAFator) and finally lineage determination (pangolin). 
                         """, style={"font-size": 16}),
            html.Br(),
            html.Div(
                html.Span(
                    children=[
                        get_mini_container(
                            title="Samples",
                            value=print_number(count_samples)
                        ),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="First sample",
                            value=print_date(date_of_first_sample)
                        ),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="Latest sample",
                            value=print_date(date_of_most_recent_sample)
                        ),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="Unique mutations",
                            value=print_number(count_variants)
                        ),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="Mutation calls",
                            value=print_number(count_variant_observations)
                        )
                    ]
                )
            ),
            html.Br(),
            dcc.Markdown("""
                There are four exclusion criteria:
                * Non-human host samples 
                * Non Illumina sequenced samples (ie: a large proportion of nanopore sequenced samples are excluded)
                * Horizontal coverage below 20 % of the reference genome
                * Mean mapping quality below 10 or mean base call quality below 10
                
                Mutations are classified according to their VAF into clonal mutations (0.8 <= VAF <= 1.0), 
                intrahost mutations (0.05 <= VAF < 0.8) and finally low frequency mutations (VAF < 0.05).
                """, style={"font-size": 16}),
        ])


def get_dataset_ena_tab_graphs(queries: Queries):
    return html.Div(
        children=[
            html.Br(),
            html.Div(children=[
                html.Div(children=get_plot_library_strategies(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Div(children=get_plot_number_reads(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Br(),
                html.Div(children=get_plot_coverage(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Div(children=get_plot_qualities(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
            ]),
        ])


def get_plot_library_strategies(queries: Queries):
    data = get_data_library_strategies(queries)
    fig = px.bar(data_frame=data, x="library_strategy", y="count", color="library_layout",
                 color_discrete_map=LIBRARY_LAYOUT_COLOR_MAP, barmode="group")
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "Num. of samples"},
        xaxis={'title': None},
        legend={'title': None}
    )
    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Library strategies**
        
        The number of samples for each library preparation strategy.
        Only Illumina samples are included, other technologies are excluded from CoVigator at the moment.
        """)
    ]


def get_data_library_strategies(queries):
    sql_query = """
    select count(*) as count, library_strategy, library_layout
    from {table} 
    where status = '{status}'
    group by library_strategy, library_layout""".format(
        table=SAMPLE_ENA_TABLE_NAME,
        status=JobStatus.FINISHED.name
    )
    data = pd.read_sql_query(sql_query, queries.session.bind)
    return data


def get_plot_number_reads(queries: Queries):
    data = get_data_number_reads(queries)

    fig = get_scatter_with_marginals(
        data=data,
        x="num_reads_after",
        y="num_reads_before",
        x_title="Num. of reads after trimming",
        y_title="Reported num. of reads")

    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Number of reads**
       
        The reported number of reads corresponds to the number reported in the ENA API.
        The number of reads after trimming corresponds to the number of reads after running fastp on the FASTQ files.
        No samples are excluded based on the number of reads.
        """)
    ]


def get_data_number_reads(queries):
    query = queries.session.query(
        SampleEna.run_accession, SampleEna.num_reads.label("num_reads_after"),
        SampleEna.read_count.label("num_reads_before"), SampleEna.library_strategy)\
        .filter(SampleEna.status == JobStatus.FINISHED.name)
    data = pd.read_sql_query(query.statement, queries.session.bind)
    return data


def get_scatter_with_marginals(data, x, y, x_title, y_title):
    fig = go.FigureWidget()
    for k, v in LIBRARY_STRATEGY_COLOR_MAP.items():
        fig.add_scattergl(
            x=data[data.library_strategy == k][x],
            y=data[data.library_strategy == k][y],
            mode='markers',
            marker=dict(color=v),
            name=k,
            opacity=0.5,
            hovertext=data[data.library_strategy == k].run_accession
        )
        fig.add_box(x=data[data.library_strategy == k][x], yaxis='y2', marker=dict(color=v),
                    showlegend=False)
        fig.add_box(y=data[data.library_strategy == k][y], xaxis='x2', marker=dict(color=v),
                    showlegend=False)
    fig.layout = dict(
        template=TEMPLATE,
        margin=MARGIN,
        yaxis={'title': y_title,
               'domain': [0, 0.85], 'showgrid': True, 'zeroline': False, 'type': 'log'},
        xaxis={'title': x_title,
               'domain': [0, 0.85], 'showgrid': True, 'zeroline': False, 'type': 'log'},
        legend={'title': None, 'itemsizing': 'constant'},
        bargap=0,
        xaxis2={'domain': [0.85, 1], 'showgrid': False, 'zeroline': False, 'visible': False, 'showticklabels': False},
        yaxis2={'domain': [0.85, 1], 'showgrid': False, 'zeroline': False, 'visible': False, 'showticklabels': False},
        hovermode='closest',
    )
    return fig


def get_plot_coverage(queries: Queries):
    data = get_data_coverage(queries)

    fig = get_scatter_with_marginals(
        data=data,
        x="mean_depth",
        y="coverage",
        x_title="Mean depth of coverage",
        y_title="Genome coverage (%)")

    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Coverage**

        The percentage of the genome covered by at least one read versus the mean depth of coverage.
        Samples covering less than 20 % of the genome are excluded.
        """)
    ]


def get_data_coverage(queries):
    query = queries.session.query(
        SampleEna.run_accession, SampleEna.library_strategy, SampleEna.coverage, SampleEna.mean_depth) \
        .filter(SampleEna.status == JobStatus.FINISHED.name)
    data = pd.read_sql_query(query.statement, queries.session.bind)
    return data


def get_plot_qualities(queries: Queries):
    data = get_data_qualities(queries)

    fig = get_scatter_with_marginals(
        data=data,
        x="mean_base_quality",
        y="mean_mapping_quality",
        x_title="Mean BCQ",
        y_title="Mean MQ")

    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Mean mapping quality (MQ) and Base Call Quality (BCQ)**
        
        Samples covering with either mapping quality < 10 or base call quality < 10 are excluded
        """)
    ]


def get_data_qualities(queries):
    query = queries.session.query(
        SampleEna.run_accession, SampleEna.library_strategy, SampleEna.mean_mapping_quality,
        SampleEna.mean_base_quality)\
        .filter(SampleEna.status == JobStatus.FINISHED.name)
    data = pd.read_sql_query(query.statement, queries.session.bind)
    return data


def get_plot_clonal_variants(queries: Queries):
    data = get_data_clonal_variants(queries)
    fig = px.scatter(data_frame=data, color="variant_type", y="vaf", x="dp",
                     log_x=True, log_y=True, opacity=0.5, hover_name="variant_id",
                     marginal_x='box', marginal_y='box',
                     color_discrete_map=VARIANT_TYPE_COLOR_MAP)
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "VAF"},
        xaxis={'title': "DP"},
        legend={'title': None}
    )
    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Variant Allele Frequency (VAF) and Depth of Coverage (DP)**

        Variants with VAF >= 0.8 are considered clonal 
        """)
    ]


def get_data_clonal_variants(queries):
    query = queries.session.query(
        VariantObservation.variant_id, VariantObservation.vaf, VariantObservation.dp, VariantObservation.variant_type)
    data = pd.read_sql_query(query.statement, queries.session.bind)
    return data


def get_dataset_ena_tab_left_bar(queries: Queries):
    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            dcc.Markdown("""Select one or more genes"""),
            dcc.Dropdown(
                id="sample-dropdown",
                options=[{'label': g.name, 'value': g.name} for g in queries.get_genes()],
                value=None,
                multi=True
            )
        ])



from covigator.database.model import VariantCooccurrence, DataSource
from covigator.pipeline.cooccurrence_matrix import CooccurrenceMatrix
from covigator.tests.unit_tests.abstract_test import AbstractTest
from covigator.tests.unit_tests.mocked import get_mocked_ena_sample, get_mocked_variant, get_mocked_variant_observation


class CooccurrenceMatrixTests(AbstractTest):

    NUM_UNIQUE_VARIANTS = 10
    NUM_SAMPLES = 10
    NUM_VARIANT_OBSERVATIONS_PER_SAMPLE = 5

    def setUp(self) -> None:
        # mocks some unique variants
        self.mocked_variants = [get_mocked_variant(self.faker) for _ in range(self.NUM_UNIQUE_VARIANTS)]
        self.session.add_all(self.mocked_variants)
        self.session.commit()

        variant_observations = []
        self.samples = []
        for _ in range(self.NUM_SAMPLES):
            sample_ena = get_mocked_ena_sample(self.faker)
            self.session.add(sample_ena)
            self.session.commit()
            self.samples.append(sample_ena)
            # mocks observed variants
            for vo in self.faker.random_elements(
                    self.mocked_variants, length=self.NUM_VARIANT_OBSERVATIONS_PER_SAMPLE, unique=True):
                variant_observations.append(get_mocked_variant_observation(sample_ena, vo))
            self.session.add_all(variant_observations)
            self.session.commit()

    def test_missing_sample(self):
        self.assertRaises(
            AssertionError,
            CooccurrenceMatrix().compute,
            run_accession=None,
            source=None,
            session=self.session)

    def test_missing_session(self):
        self.assertRaises(
            AssertionError,
            CooccurrenceMatrix().compute,
            run_accession="12345",
            source=DataSource.ENA,
            session=None)

    def test_non_existing_sample_does_not_add_new_entries(self):
        count = self.session.query(VariantCooccurrence).count()
        self.assertEqual(count, 0)
        CooccurrenceMatrix().compute(run_accession="12345", source=DataSource.ENA.name, session=self.session)
        self.session.commit()
        count = self.session.query(VariantCooccurrence).count()
        self.assertEqual(count, 0)

    def test_one_existing_sample(self):
        count = self.session.query(VariantCooccurrence).count()
        self.assertEqual(count, 0)
        CooccurrenceMatrix().compute(
            run_accession=self.samples[0].run_accession, source=DataSource.ENA.name, session=self.session)
        self.session.commit()
        count = self.session.query(VariantCooccurrence).count()
        # size of matrix = n*(n-1)/2 + n (one half of matrix + diagonal)
        self.assertEqual(
            count, (self.NUM_VARIANT_OBSERVATIONS_PER_SAMPLE * (self.NUM_VARIANT_OBSERVATIONS_PER_SAMPLE - 1) / 2) +
                   self.NUM_VARIANT_OBSERVATIONS_PER_SAMPLE)

        self._assert_cooccurrence_matrix(assert_cooccurrence=False)

    def test_all_samples(self):
        count = self.session.query(VariantCooccurrence).count()
        self.assertEqual(count, 0)
        for s in self.samples:
            CooccurrenceMatrix().compute(run_accession=s.run_accession, source=DataSource.ENA.name, session=self.session)
            self.session.commit()
        count = self.session.query(VariantCooccurrence).count()
        self.assertLess(
            count,
            self.NUM_SAMPLES * self.NUM_VARIANT_OBSERVATIONS_PER_SAMPLE *
            (self.NUM_VARIANT_OBSERVATIONS_PER_SAMPLE - 1) / 2)
        self._assert_cooccurrence_matrix()

    def _assert_cooccurrence_matrix(self, assert_cooccurrence=True):
        variant_cooccurrences = self.session.query(VariantCooccurrence).all()
        found_cooccurrent_variant = False
        for vo in variant_cooccurrences:
            position_one = int(vo.variant_id_one.split(":")[0])
            position_two = int(vo.variant_id_two.split(":")[0])
            self.assertLessEqual(position_one, position_two)
            self.assertLessEqual(vo.count, self.NUM_SAMPLES)
            self.assertGreater(vo.count, 0)
            if vo.count > 1:
                found_cooccurrent_variant = True
        if assert_cooccurrence:
            self.assertTrue(found_cooccurrent_variant)




from unittest import TestCase
from faker import Faker
from logzero import logger
from covigator.database.database import Database
from covigator.database.model import VariantCooccurrence, SampleEna, VariantObservation, \
    SampleGisaid, SubclonalVariantObservation, Variant, PrecomputedVariantAbundanceHistogram, PrecomputedTableCounts, \
    PrecomputedSynonymousNonSynonymousCounts, PrecomputedOccurrence, PrecomputedAnnotation, PrecomputedIndelLength, \
    PrecomputedSubstitutionsCounts, PrecomputedVariantsPerSample, Log, GisaidVariant, GisaidVariantObservation, \
    LowFrequencyVariantObservation, LowFrequencyVariant, SubclonalVariant, PrecomputedVariantsPerLineage, \
    GisaidVariantCooccurrence
from covigator.database.queries import Queries
from covigator.tests.unit_tests.faked_objects import FakeConfiguration


class AbstractTest(TestCase):

    @classmethod
    def setUpClass(cls):
        """On inherited classes, run our `setUp` method"""
        # Inspired via http://stackoverflow.com/questions/1323455/python-unit-test-with-base-and-sub-class/17696807#17696807
        if cls is not AbstractTest and cls.setUp is not AbstractTest.setUp:
            orig_setUp = cls.setUp

            def setUpOverride(self, *args, **kwargs):
                AbstractTest.setUp(self)
                return orig_setUp(self, *args, **kwargs)

            cls.setUp = setUpOverride

    def setUp(self) -> None:
        self.config = FakeConfiguration()
        self.database = Database(test=True, config=self.config)
        self.session = self.database.get_database_session()
        self.queries = Queries(session=self.session)
        self.faker = Faker()
        self._clean_test_database()

    def tearDown(self) -> None:
        logger.info("Cleaning the database")
        self._clean_test_database()

    def _clean_test_database(self):
        try:
            self._clean_table(GisaidVariantObservation)
            self._clean_table(VariantObservation)
            self._clean_table(SubclonalVariantObservation)
            self._clean_table(LowFrequencyVariantObservation)
            self._clean_table(SampleEna)
            self._clean_table(SampleGisaid)
            self._clean_table(VariantCooccurrence)
            self._clean_table(GisaidVariantCooccurrence)
            self._clean_table(Variant)
            self._clean_table(SubclonalVariant)
            self._clean_table(LowFrequencyVariant)
            self._clean_table(GisaidVariant)
            self._clean_table(PrecomputedVariantsPerSample)
            self._clean_table(PrecomputedSubstitutionsCounts)
            self._clean_table(PrecomputedIndelLength)
            self._clean_table(PrecomputedAnnotation)
            self._clean_table(PrecomputedOccurrence)
            self._clean_table(PrecomputedSynonymousNonSynonymousCounts)
            self._clean_table(PrecomputedTableCounts)
            self._clean_table(PrecomputedVariantAbundanceHistogram)
            self._clean_table(PrecomputedVariantsPerLineage)
            self._clean_table(Log)
        except Exception as e:
            logger.error("Error cleaning the database")
            logger.exception(e)

    def _clean_table(self, clazz):
        self.session.query(clazz).delete()
        self.session.commit()




VERSION = "v0.7.1"
ANALYSIS_PIPELINE_VERSION = "v0.9.3"

MISSENSE_VARIANT = "missense_variant"
SYNONYMOUS_VARIANT = "synonymous_variant"
INFRAME_DELETION = "inframe_deletion"
INFRAME_INSERTION = "inframe_insertion"
DISRUPTIVE_INFRAME_DELETION = "disruptive_inframe_deletion"
CONSERVATIVE_INFRAME_DELETION = "conservative_inframe_deletion"
DISRUPTIVE_INFRAME_INSERTION = "disruptive_inframe_insertion"
CONSERVATIVE_INFRAME_INSERTION = "conservative_inframe_insertion"
